import glob
import os

import numpy as np
from sklearn.datasets import load_breast_cancer
from sklearn.decomposition import PCA
from sklearn.pipeline import Pipeline as SKPipeline
from sklearn.preprocessing import StandardScaler
from sklearn.svm import SVC
from sklearn.tree import DecisionTreeClassifier

from photonai.base import PipelineElement, Preprocessing, Stack, Switch, Branch, \
    CallbackElement
from photonai.base.cache_manager import CacheManager
from photonai.base.photon_pipeline import PhotonPipeline
from photonai.helper.dummy_elements import DummyYAndCovariatesTransformer
from photonai.helper.photon_base_test import PhotonBaseTest


# assertEqual(a, b) 	a == b
# assertNotEqual(a, b) 	a != b
# assertTrue(x) 	bool(x) is True
# assertFalse(x) 	bool(x) is False
# assertIs(a, b) 	a is b 	3.1
# assertIsNot(a, b) 	a is not b 	3.1
# assertIsNone(x) 	x is None 	3.1
# assertIsNotNone(x) 	x is not None 	3.1
# assertIn(a, b) 	a in b 	3.1
# assertNotIn(a, b) 	a not in b 	3.1
# assertIsInstance(a, b) 	isinstance(a, b) 	3.2
# assertNotIsInstance(a, b) 	not isinstance(a, b) 	3.2


class PipelineTests(PhotonBaseTest):

    @classmethod
    def setUpClass(cls) -> None:
        cls.file = __file__
        super(PipelineTests, cls).setUpClass()

    def setUp(self):
        super(PipelineTests, self).setUp()
        self.X, self.y = load_breast_cancer(return_X_y=True)

        # Photon Version
        self.p_pca = PipelineElement("PCA", {}, random_state=3)
        self.p_svm = PipelineElement("SVC", {}, random_state=3)
        self.p_ss = PipelineElement("StandardScaler", {})
        self.p_dt = PipelineElement("DecisionTreeClassifier", random_state=3)

        dummy_element = DummyYAndCovariatesTransformer()
        self.dummy_photon_element = PipelineElement.create("DummyTransformer", dummy_element, {})

        self.sk_pca = PCA(random_state=3)
        self.sk_svc = SVC(random_state=3)
        self.sk_ss = StandardScaler()
        self.sk_dt = DecisionTreeClassifier(random_state=3)

    def test_regular_use(self):

        photon_pipe = PhotonPipeline([("PCA", self.p_pca), ("SVC", self.p_svm)])
        photon_pipe.fit(self.X, self.y)

        photon_transformed_X, _, _ = photon_pipe.transform(self.X)
        photon_predicted_y = photon_pipe.predict(self.X)

        # the element is given by reference, so it should be fitted right here
        photon_ref_transformed_X, _, _ = self.p_pca.transform(self.X)
        photon_ref_predicted_y = self.p_svm.predict(photon_ref_transformed_X)

        self.assertTrue(np.array_equal(photon_transformed_X, photon_ref_transformed_X))
        self.assertTrue(np.array_equal(photon_predicted_y, photon_ref_predicted_y))

        sk_pipe = SKPipeline([('PCA', self.sk_pca), ("SVC", self.sk_svc)])
        sk_pipe.fit(self.X, self.y)

        sk_predicted_y = sk_pipe.predict(self.X)
        self.assertTrue(np.array_equal(photon_predicted_y, sk_predicted_y))

        # sklearn pipeline does not offer a transform function
        # sk_transformed_X = sk_pipe.transform(X)
        # self.assertTrue(np.array_equal(photon_transformed_X, sk_transformed_X))

    def test_add_preprocessing(self):
        my_preprocessing = Preprocessing()
        my_preprocessing += PipelineElement('LabelEncoder')
        photon_pipe = PhotonPipeline([("PCA", self.p_pca), ("SVC", self.p_svm)])
        photon_pipe.add_preprocessing(my_preprocessing)

        self.assertEqual(len(photon_pipe.named_steps), 3)
        first_element = photon_pipe.elements[0][1]
        self.assertTrue(first_element == my_preprocessing)
        self.assertTrue(photon_pipe.named_steps['Preprocessing'] == my_preprocessing)

    def test_no_estimator(self):

        no_estimator_pipe = PhotonPipeline([("StandardScaler", self.p_ss), ("PCA", self.p_pca)])
        no_estimator_pipe.fit(self.X, self.y)
        photon_no_estimator_transform, _, _ = no_estimator_pipe.transform(self.X)
        photon_no_estimator_predict = no_estimator_pipe.predict(self.X)

        self.assertTrue(np.array_equal(photon_no_estimator_predict, photon_no_estimator_transform))

        self.sk_ss.fit(self.X)
        standardized_data = self.sk_ss.transform(self.X)
        self.sk_pca.fit(standardized_data)
        pca_data = self.sk_pca.transform(standardized_data)

        self.assertTrue(np.array_equal(photon_no_estimator_transform, pca_data))
        self.assertTrue(np.array_equal(photon_no_estimator_predict, pca_data))

    def test_y_and_covariates_transformation(self):

        X = np.ones((200, 50))
        y = np.ones((200,)) + 2
        kwargs = {'sample1': np.ones((200, 5))}

        photon_pipe = PhotonPipeline([("DummyTransformer", self.dummy_photon_element)])

        # if y is none all y transformer should be ignored
        Xt2, yt2, kwargst2 = photon_pipe.transform(X, None, **kwargs)
        self.assertTrue(np.array_equal(Xt2, X))
        self.assertTrue(np.array_equal(yt2, None))
        self.assertTrue(np.array_equal(kwargst2, kwargs))

        # if y is given, all y transformers should be working
        Xt, yt, kwargst = photon_pipe.transform(X, y, **kwargs)

        # assure that data is delivered to element correctly
        self.assertTrue(np.array_equal(X, self.dummy_photon_element.base_element.X))
        self.assertTrue(np.array_equal(y, self.dummy_photon_element.base_element.y))
        self.assertTrue(np.array_equal(kwargs["sample1"], self.dummy_photon_element.base_element.kwargs["sample1"]))

        # assure that data is transformed correctly
        self.assertTrue(np.array_equal(Xt, X - 1))
        self.assertTrue(np.array_equal(yt, y + 1))
        self.assertTrue("sample1_edit" in kwargst)
        self.assertTrue(np.array_equal(kwargst["sample1_edit"], kwargs["sample1"] + 5))

    def test_predict_with_training_flag(self):
        # manually edit labels
        sk_pipe = SKPipeline([("SS", self.sk_ss), ("SVC", self.sk_svc)])
        y_plus_one = self.y + 1
        sk_pipe.fit(self.X, y_plus_one)
        sk_pred = sk_pipe.predict(self.X)

        # edit labels during pipeline
        p_pipe = PhotonPipeline([("SS", self.p_ss), ("YT", self.dummy_photon_element), ("SVC", self.p_svm)])
        p_pipe.fit(self.X, self.y)
        p_pred = p_pipe.predict(self.X)

        sk_standardized_X = self.sk_ss.transform(self.X)
        input_of_y_transformer = self.dummy_photon_element.base_element.X
        self.assertTrue(np.array_equal(sk_standardized_X, input_of_y_transformer))

        self.assertTrue(np.array_equal(sk_pred, p_pred))

    def test_inverse_tansform(self):
        # simple pipe
        sk_pipe = SKPipeline([("SS", self.sk_ss), ("PCA", self.sk_pca)])
        sk_pipe.fit(self.X, self.y)
        sk_transform = sk_pipe.transform(self.X)
        sk_inverse_transformed = sk_pipe.inverse_transform(sk_transform)

        photon_pipe = PhotonPipeline([("SS", self.p_ss), ("PCA", self.p_pca)])
        photon_pipe.fit(self.X, self.y)
        p_transform, _, _ = photon_pipe.transform(self.X)
        p_inverse_transformed, _, _ = photon_pipe.inverse_transform(p_transform)

        self.assertTrue(np.array_equal(sk_inverse_transformed, p_inverse_transformed))

        # now including stack
        stack = Stack('stack', [self.p_pca])
        stack_pipeline = PhotonPipeline([("stack", stack), ('StandardScaler', PipelineElement('StandardScaler')),
                                         ('LinearSVC', PipelineElement('LinearSVC'))])
        stack_pipeline.fit(self.X, self.y)
        feature_importances = stack_pipeline.feature_importances_
        inversed_data, _, _ = stack_pipeline.inverse_transform(feature_importances)
        self.assertEqual(inversed_data.shape[1], self.X.shape[1])

    # Todo: add tests for kwargs

    def test_predict_proba(self):

        sk_pipe = SKPipeline([("SS", self.sk_ss), ("SVC", self.sk_dt)])
        sk_pipe.fit(self.X, self.y)
        sk_proba = sk_pipe.predict_proba(self.X)

        photon_pipe = PhotonPipeline([("SS", self.p_ss), ("SVC", self.p_dt)])
        photon_pipe.fit(self.X, self.y)
        photon_proba = photon_pipe.predict_proba(self.X)

        self.assertTrue(np.array_equal(sk_proba, photon_proba))

    def test_copy_me(self):
        switch = Switch("my_copy_switch")
        switch += PipelineElement("StandardScaler")
        switch += PipelineElement("RobustScaler", test_disabled=True)

        stack = Stack("RandomStack")
        stack += PipelineElement("SVC")
        branch = Branch('Random_Branch')
        pca_hyperparameters = {'n_components': [5, 10]}
        branch += PipelineElement("PCA", hyperparameters=pca_hyperparameters)
        branch += PipelineElement("DecisionTreeClassifier")
        stack += branch

        photon_pipe = PhotonPipeline([("SimpleImputer", PipelineElement("SimpleImputer")),
                                      ("my_copy_switch", switch),
                                      ('RandomStack', stack),
                                      ('Callback1', CallbackElement('tmp_callback', np.mean)),
                                      ("PhotonVotingClassifier", PipelineElement("PhotonVotingClassifier"))])

        copy_of_the_pipe = photon_pipe.copy_me()

        self.assertEqual(photon_pipe.random_state, copy_of_the_pipe.random_state)
        self.assertTrue(len(copy_of_the_pipe.elements) == 5)
        self.assertTrue(copy_of_the_pipe.elements[2][1].name == "RandomStack")
        self.assertTrue(copy_of_the_pipe.named_steps["my_copy_switch"].elements[1].test_disabled)
        self.assertDictEqual(copy_of_the_pipe.elements[2][1].elements[1].elements[0].hyperparameters,
                             {"PCA__n_components": [5, 10]})
        self.assertTrue(isinstance(copy_of_the_pipe.elements[3][1], CallbackElement))
        self.assertTrue(copy_of_the_pipe.named_steps["tmp_callback"].delegate_function == np.mean)

    def test_random_state(self):
        photon_pipe = PhotonPipeline([("SS", self.p_ss), ("PCA", PipelineElement('PCA')), ("SVC", self.p_dt)])
        photon_pipe.random_state = 666
        photon_pipe.fit(self.X, self.y)
        self.assertEqual(self.p_dt.random_state, photon_pipe.random_state)
        self.assertEqual(photon_pipe.elements[1][-1].random_state, photon_pipe.random_state)
        self.assertEqual(self.p_dt.random_state, 666)


class CacheManagerTests(PhotonBaseTest):

    @classmethod
    def setUpClass(cls) -> None:
        cls.file = __file__
        super(CacheManagerTests, cls).setUpClass()

    def setUp(self):
        super(CacheManagerTests, self).setUp()

        self.cache_man = CacheManager("123353423434", self.cache_folder_path)
        self.X, self.y, self.kwargs = np.array([1, 2, 3, 4, 5]), np.array([1, 2, 3, 4, 5]), \
                                      {'covariates': [9, 8, 7, 6, 5]}

        self.config1 = {'PCA__n_components': 5,
                        'SVC__C': 3,
                        'SVC__kernel': 'rbf'}
        self.item_names = ["StandardScaler", "PCA", "SVC"]

        self.config2 = {'PCA__n_components': 20,
                        'SVC__C': 1,
                        'SVC__kernel': 'linear'}

    def test_find_relevant_configuration_items(self):
        self.cache_man.prepare(pipe_elements=self.item_names, X=self.X, config=self.config1)
        relevant_items = {'PCA__n_components': 5}
        relevant_items_hash = hash(frozenset(relevant_items.items()))
        new_hash = self.cache_man._find_config_for_element("PCA")
        self.assertEqual(relevant_items_hash, new_hash)

    def test_empty_config(self):
        self.cache_man.prepare(pipe_elements=self.item_names, X=self.X, config={})
        relevant_items_hash = hash(frozenset({}.items()))
        new_hash = self.cache_man._find_config_for_element("PCA")
        self.assertEqual(relevant_items_hash, new_hash)

    def test_initial_transformation(self):
        self.cache_man.prepare(pipe_elements=self.item_names, config=self.config1)
        result = self.cache_man.load_cached_data("PCA")
        self.assertEqual(result, None)

    def test_check_cache(self):
        self.cache_man.prepare(pipe_elements=self.item_names, config=self.config1)
        self.assertFalse(self.cache_man.check_cache("PCA"))
        self.cache_man.save_data_to_cache("PCA", (self.X, self.y, self.kwargs))
        self.assertTrue(self.cache_man.check_cache("PCA"))

    def test_key_hash_equal(self):
        self.cache_man.prepare(pipe_elements=self.item_names, config=self.config1)
        generator_1 = self.cache_man.generate_cache_key("PCA")
        generator_2 = self.cache_man.generate_cache_key("PCA")
        self.assertEqual(generator_1, generator_2)

    def test_saving_and_loading_transformation(self):
        self.cache_man.prepare(pipe_elements=self.item_names, config=self.config1)
        self.cache_man.save_data_to_cache("PCA", (self.X, self.y, self.kwargs))

        self.assertTrue(len(self.cache_man.cache_index) == 1)
        for hash_key, cache_file in self.cache_man.cache_index.items():
            self.assertTrue(os.path.isfile(cache_file))

        result = self.cache_man.load_cached_data("PCA")
        self.assertTrue(result is not None)
        X_loaded, y_loaded, kwargs_loaded = result[0], result[1], result[2]
        self.assertTrue(np.array_equal(self.X, X_loaded))
        self.assertTrue(np.array_equal(self.y, y_loaded))
        self.assertTrue(np.array_equal(self.kwargs['covariates'], kwargs_loaded['covariates']))

    def test_clearing_folder(self):
        self.cache_man.clear_cache()
        self.assertTrue(len(glob.glob(os.path.join(self.cache_man.cache_folder, "*.p"))) == 0)


class CachedPhotonPipelineTests(PhotonBaseTest):

    @classmethod
    def setUpClass(cls) -> None:
        cls.file = __file__
        super(CachedPhotonPipelineTests, cls).setUpClass()

    def setUp(self):
        super(CachedPhotonPipelineTests, self).setUp()
        # Photon Version
        ss = PipelineElement("StandardScaler", {})
        pca = PipelineElement("PCA", {'n_components': [3, 10, 50]}, random_state=3)
        svm = PipelineElement("SVC", {'kernel': ['rbf', 'linear']}, random_state=3)

        self.pipe = PhotonPipeline([('StandardScaler', ss),
                                    ('PCA', pca),
                                    ('SVC', svm)])

        self.pipe.caching = True
        self.pipe.fold_id = "12345643463434"
        CacheManager.clear_cache_files(self.cache_folder_path)
        self.pipe.cache_folder = self.cache_folder_path

        self.config1 = {'PCA__n_components': 4,
                        'SVC__C': 3,
                        'SVC__kernel': 'rbf'}

        self.config2 = {'PCA__n_components': 7,
                        'SVC__C': 1,
                        'SVC__kernel': 'linear'}

        self.X, self.y = load_breast_cancer(return_X_y=True)

    def test_group_caching(self):

        # transform one config
        self.pipe.set_params(**self.config1)
        self.pipe.fit(self.X, self.y)
        X_new, y_new, kwargs_new = self.pipe.transform(self.X, self.y)
        # one result should be cached ( one standard scaler output + one pca output)
        self.assertTrue(len(glob.glob(os.path.join(self.pipe.cache_folder, "*.p"))) == 2)

        # transform second config
        self.pipe.set_params(**self.config2)
        self.pipe.fit(self.X, self.y)
        X_config2, y_config2, kwargs_config2 = self.pipe.transform(self.X, self.y)
        # two results should be cached ( one standard scaler output (config hasn't changed)
        # + two pca outputs  )
        self.assertTrue(len(glob.glob(os.path.join(self.pipe.cache_folder, "*.p"))) == 3)

        # now transform with config 1 again, results should be loaded
        self.pipe.set_params(**self.config1)
        self.pipe.fit(self.X, self.y)
        X_2, y_2, kwargs_2 = self.pipe.transform(self.X, self.y)
        self.assertTrue(np.array_equal(X_new, X_2))
        self.assertTrue(np.array_equal(y_new, y_2))
        self.assertTrue(np.array_equal(kwargs_new, kwargs_2))

        # results should be the same as when caching is deactivated
        self.pipe.caching = False
        self.pipe.set_params(**self.config1)
        self.pipe.fit(self.X, self.y)
        X_uc, y_uc, kwargs_uc = self.pipe.transform(self.X, self.y)
        self.assertTrue(np.array_equal(X_uc, X_2))
        self.assertTrue(np.array_equal(y_uc, y_2))
        self.assertTrue(np.array_equal(kwargs_uc, kwargs_2))

    def test_empty_hyperparameters(self):
        # test if one can use it when only default parameters are given and hyperparameter space is empty
        self.pipe.set_params(**{})
        self.pipe.fit(self.X, self.y)
        X_new, y_new, kwargs_new = self.pipe.transform(self.X, self.y)
        # one result should be cached ( one standard scaler output + one pca output )
        self.assertTrue(len(glob.glob(os.path.join(self.pipe.cache_folder, "*.p"))) == 2)

        self.pipe.set_params(**{})
        self.pipe.fit(self.X, self.y)
        X_new2, y_new2, kwargs_new2 = self.pipe.transform(self.X, self.y)
        # assert nothing happened in the cache folder
        self.assertTrue(len(glob.glob(os.path.join(self.pipe.cache_folder, "*.p"))) == 2)
        self.assertTrue(np.array_equal(X_new, X_new2))
        self.assertTrue(np.array_equal(y_new, y_new2))
        self.assertTrue(np.array_equal(kwargs_new, kwargs_new2))




from sklearn.datasets import load_boston
from sklearn.model_selection import KFold

from photonai.base import Hyperpipe, PipelineElement
from photonai.optimization import IntegerRange, FloatRange


my_pipe = Hyperpipe('basic_regression_pipe',
                    optimizer='random_search',
                    optimizer_params={'n_configurations': 25},
                    metrics=['mean_squared_error', 'mean_absolute_error', 'explained_variance'],
                    best_config_metric='mean_squared_error',
                    outer_cv=KFold(n_splits=3, shuffle=True),
                    inner_cv=KFold(n_splits=3, shuffle=True),
                    verbosity=1,
                    project_folder='./tmp/')

my_pipe += PipelineElement('SimpleImputer')
my_pipe += PipelineElement('StandardScaler')

my_pipe += PipelineElement('LassoFeatureSelection',
                           hyperparameters={'percentile': [0.1, 0.2, 0.3],
                                            'alpha': FloatRange(0.5, 5)})

my_pipe += PipelineElement('RandomForestRegressor',
                           hyperparameters={'n_estimators': IntegerRange(10, 50)})

# load data and train
X, y = load_boston(return_X_y=True)
my_pipe.fit(X, y)



from sklearn.datasets import load_breast_cancer
from sklearn.model_selection import KFold

from photonai.base import Hyperpipe, PipelineElement, Switch
from photonai.optimization import IntegerRange

# GET DATA
X, y = load_breast_cancer(return_X_y=True)

# CREATE HYPERPIPE
my_pipe = Hyperpipe('basic_switch_pipe',
                    optimizer='random_grid_search',
                    optimizer_params={'n_configurations': 15},
                    metrics=['accuracy', 'precision', 'recall'],
                    best_config_metric='accuracy',
                    outer_cv=KFold(n_splits=3),
                    inner_cv=KFold(n_splits=5),
                    verbosity=1,
                    project_folder='./tmp/')

# Transformer Switch
my_pipe += Switch('StandardizationSwitch',
                  [PipelineElement('StandardScaler'),
                   PipelineElement('MinMaxScaler')])

# Estimator Switch
svm = PipelineElement('SVC',
                      hyperparameters={'kernel': ['rbf', 'linear']})

tree = PipelineElement('DecisionTreeClassifier',
                       hyperparameters={'min_samples_split': IntegerRange(2, 5),
                                        'min_samples_leaf': IntegerRange(1, 5),
                                        'criterion': ['gini', 'entropy']})

my_pipe += Switch('EstimatorSwitch', [svm, tree])

my_pipe.fit(X, y)



# cnn model with PHOTONAI
# content by J. Brownlee:
# https://machinelearningmastery.com/cnn-models-for-human-activity-recognition-time-series-classification/
# HAR-Dataset: https://archive.ics.uci.edu/ml/datasets/human+activity+recognition+using+smartphones
# required file: data.py from examples/neural_network

from keras.utils import data_utils
from keras.models import Sequential
from keras.layers import Dense, Flatten, Dropout
from keras.layers.convolutional import Conv1D, MaxPooling1D
from sklearn.model_selection import StratifiedShuffleSplit

from examples.neural_networks.dataset import load_har

from photonai.base import Hyperpipe, PipelineElement
from photonai.optimization import Categorical

dataset_path = data_utils.get_file(
    fname='UCI HAR Dataset.zip',
    origin='https://archive.ics.uci.edu/ml/machine-learning-databases/00240/UCI%20HAR%20Dataset.zip',
    file_hash='53e099237392e0b9602f8c38f578bd8f',
    hash_algorithm='md5',
    cache_subdir='photonai_datasets',
    extract=True,
    archive_format='zip'
)

X, y = load_har(prefix=dataset_path.replace('.zip', ''))

n_timesteps, n_features, n_outputs = X.shape[1], X.shape[2], 6
model = Sequential()
model.add(Conv1D(filters=64, kernel_size=3, activation='relu', input_shape=(n_timesteps, n_features)))
model.add(Conv1D(filters=64, kernel_size=3, activation='relu'))
model.add(Dropout(0.5))
model.add(MaxPooling1D(pool_size=2))
model.add(Flatten())
model.add(Dense(100, activation='relu'))
model.add(Dense(n_outputs, activation='softmax'))
model.compile(loss='categorical_crossentropy', optimizer='adam', metrics=['accuracy'])

# DESIGN YOUR PIPELINE
my_pipe = Hyperpipe('cnn_keras_multiclass_pipe',
                    optimizer='grid_search',
                    optimizer_params={},
                    metrics=['accuracy'],
                    best_config_metric='accuracy',
                    outer_cv=StratifiedShuffleSplit(n_splits=3, test_size=0.2),
                    inner_cv=StratifiedShuffleSplit(n_splits=2, test_size=0.2),
                    verbosity=1,
                    project_folder='./tmp/')

my_pipe += PipelineElement('KerasBaseClassifier',
                           hyperparameters={'epochs': Categorical([10, 20])},
                           verbosity=1,
                           model=model)

# NOW TRAIN YOUR PIPELINE
my_pipe.fit(X, y)



import uuid
import numpy as np
from sklearn.model_selection import GroupKFold, GroupShuffleSplit, LeaveOneGroupOut, StratifiedKFold, \
    StratifiedShuffleSplit, ShuffleSplit

from photonai.photonlogger.logger import logger
from photonai.processing.cross_validation import StratifiedKFoldRegression


class FoldInfo:
    """Fold Info.

    Provides all information of a fold.
    This guarantees the mapping between ID,
    number and indices for training and testing.

    Parameters
    ----------
    fold_id: default=None
        UUID as a unique indicator.

    fold_nr: int, default=0
        Position of Fold in cross-validation.

    train_indices: list, default=None
        List of all indices defined in the training set.

    test_indices: list, default=None
        List of all indices defined in the test set.

    """

    def __init__(self, fold_id=None, fold_nr: int = 0, train_indices: list = None, test_indices: list = None):
        self.fold_id = fold_id
        self.fold_nr = fold_nr
        self.train_indices = train_indices
        self.test_indices = test_indices

    @staticmethod
    def data_overview(y):
        if len(y.shape) > 1:
            # one hot encoded
            logger.info("One Hot Encoded data fold information not yet implemented")
            return {}
        else:
            unique, counts = np.unique(y, return_counts=True)
            # replacing is necessary for float inputs, because mongoDB does not allow '.'-char
            unique = [str(u).replace(".", "_") for u in unique]
            counts = [int(c) for c in counts]
            return dict(zip(unique, counts))

    @staticmethod
    def generate_folds(cv_strategy, X, y, kwargs, eval_final_performance=True, test_size=0.2):
        """Generates the training and  test set indices for the hyperparameter search.
        Returns a tuple of training and test indices.

        Parameters
        ----------
        cv_strategy
            Strategy for generating train/test sets.

        X: ndarray
            Data X to split.

        y: ndarray
            Targets to split.

        kwargs: dict
            Mostly empty. Responds only to the keyword "groups"
            and returns the corresponding value in the cv_strategy.

        eval_final_performance: bool, default=True
            See below in notes and code for the explanation.

        test_size: float, default=0.2,
            Only relevant if the final performance is performed.
            Defines the percentage size of the test set.

        Notes
        -----
        - If there is a strategy given for the outer cross validation the strategy is called to split the data
            - additionally, if a group variable and a GroupCV is passed, split data according to groups.
            - if a group variable and a StratifiedCV is passed, split data according to groups and ignore targets when
            stratifying the data.
            - if no group variable but a StratifiedCV is passed, split data according to targets.
        - If no strategy is given and eval_final_performance is True, all data is used for training.
        - If no strategy is given and eval_final_performance is False: a test set is seperated from the
          training and validation set by the parameter test_size with ShuffleSplit.

        """
        # if there is a CV Object for cross validating the hyperparameter search
        if 'groups' in kwargs.keys():
            groups = kwargs['groups']
        else:
            groups = None

        if cv_strategy is not None:
            if groups is not None and (isinstance(cv_strategy, (GroupKFold, GroupShuffleSplit, LeaveOneGroupOut))):
                try:
                    data_test_cases = cv_strategy.split(X, y, groups)
                except:
                    logger.error("Could not split data according to groups")
            elif groups is not None and (isinstance(cv_strategy, (StratifiedKFoldRegression,
                                                                  StratifiedKFold,
                                                                  StratifiedShuffleSplit))):
                try:
                    data_test_cases = cv_strategy.split(X, groups)
                except:
                    logger.error("Could not stratify data for outer cross validation according to "
                                 "group variable")
            else:
                data_test_cases = cv_strategy.split(X, y)

        # in case we do not want to divide between validation and test set
        # Re eval_final_performance:
        #         # set eval_final_performance to False because
        #         # 1. if no cv-object is given, no split is performed --> seems more logical
        #         #    than passing nothing, passing no cv-object but getting
        #         #    an 80/20 split by default
        #         # 2. if cv-object is given, split is performed but we don't peek
        #         #    into the test set --> thus we can evaluate more hp configs
        #         #    later without double dipping
        elif not eval_final_performance:
            data_test_cases = FoldInfo._yield_all_data(X)
        # the default is dividing one time into a validation and test set
        else:
            train_test_cv_object = ShuffleSplit(n_splits=1, test_size=test_size)
            data_test_cases = train_test_cv_object.split(X, y)

        fold_objects = list()
        for i, (train_indices, test_indices) in enumerate(data_test_cases):
            fold_info_obj = FoldInfo(fold_id=uuid.uuid4(),
                                     fold_nr=i+1,
                                     train_indices=train_indices,
                                     test_indices=test_indices)
            fold_objects.append(fold_info_obj)

        return fold_objects

    @staticmethod
    def _yield_all_data(X):
        """
        Helper function that iteratively returns the data stored in self.X
        Returns an iterable version of self.X
        """
        if hasattr(X, 'shape'):
            yield np.asarray(list(range(X.shape[0]))), []
        else:
            yield np.asarray(list(range(len(X)))), []



import unittest
import warnings
import numpy as np

from sklearn.datasets import load_breast_cancer
from sklearn.decomposition import PCA
from sklearn.svm import SVC
from sklearn.pipeline import Pipeline as SKPipeline
from sklearn.preprocessing import StandardScaler
from sklearn.tree import DecisionTreeClassifier

from photonai.base import PipelineElement, Switch, Stack, Branch, DataFilter, CallbackElement, Preprocessing
from photonai.base.photon_pipeline import PhotonPipeline
from photonai.helper.helper import PhotonDataHelper
from photonai.helper.dummy_elements import DummyEstimator, \
    DummyNeedsCovariatesEstimator, DummyNeedsCovariatesTransformer, DummyNeedsYTransformer, DummyTransformer, \
    DummyNeedsCovariatesAndYTransformer, DummyEstimatorNoPredict, DummyEstimatorWrongType, DummyTransformerWithPredict
from photonai.helper.photon_base_test import elements_to_dict
from photonai.optimization import GridSearchOptimizer


class PipelineElementTests(unittest.TestCase):

    def setUp(self):
        self.pca_pipe_element = PipelineElement('PCA', {'n_components': [1, 2]}, test_disabled=True, random_state=42)
        self.svc_pipe_element = PipelineElement('SVC', {'C': [0.1, 1], 'kernel': ['rbf', 'sigmoid']}, random_state=42)
        self.X, self.y = load_breast_cancer(return_X_y=True)
        self.kwargs = {'covariates': self.y}
        self.Xt = self.X + 1
        self.yt = self.y + 1
        self.kwargst = {'covariates': self.y + 1}

    def test_create_failure(self):
        with self.assertRaises(NameError):
            PipelineElement('NONSENSEName', {})

    def test_pipeline_element_create(self):
        # test name, set_disabled and base_element
        self.assertIsInstance(self.pca_pipe_element.base_element, PCA)

        # set_disabled is passed correctly
        self.assertTrue(self.pca_pipe_element.test_disabled)

        # correct name
        self.assertEqual(self.pca_pipe_element.name, 'PCA')

    def test_create_from_object(self):
        pca = PipelineElement.create('obj_name', PCA(), hyperparameters={'n_components': [1, 4]})
        self.assertIsInstance(pca.base_element, PCA)
        self.assertEqual(pca.name, 'obj_name')

        with self.assertRaises(ValueError):
            pca_2 = PipelineElement.create('obj_name', PCA, hyperparameters={'n_components': [1, 4]})

    def test_fit_and_score(self):
        tmp_pca = PCA().fit(self.X, self.y)
        self.pca_pipe_element.fit(self.X, self.y)
        self.assertEqual(self.pca_pipe_element.base_element.components_.shape, (30, 30))
        self.assertAlmostEqual(self.pca_pipe_element.base_element.components_[0, 0], tmp_pca.components_[0, 0])

        tmp_svc = SVC().fit(self.X, self.y)
        self.svc_pipe_element.fit(self.X, self.y)
        self.assertAlmostEqual(self.svc_pipe_element.base_element._intercept_[0], tmp_svc._intercept_[0])

        sk_score = tmp_svc.score(self.X, self.y)
        p_score = self.svc_pipe_element.score(self.X, self.y)
        self.assertTrue(np.array_equal(sk_score, p_score))

    def test_transform(self):
        self.pca_pipe_element.fit(self.X, self.y)

        Xt, _, _ = self.pca_pipe_element.transform(self.X)
        self.assertEqual(Xt.shape, (569, 30))
        self.assertAlmostEqual(Xt[0, 0], 1160.1425737041347)

    def test_predict(self):
        self.svc_pipe_element.fit(self.X, self.y)

        yt = self.svc_pipe_element.predict(self.X)
        self.assertEqual(yt.shape, (569,))
        self.assertEqual(yt[21], 1)

    def test_predict_proba(self):
        self.svc_pipe_element.fit(self.X, self.y)
        self.assertEqual(self.svc_pipe_element.predict_proba(self.X), None)

        gpc = PipelineElement('GaussianProcessClassifier')
        gpc.fit(self.X, self.y)
        self.assertTrue(np.array_equal(gpc.predict_proba(self.X)[0], np.asarray([0.5847072926551391, 0.4152927073448609])))

    def test_inverse_transform(self):
        Xt, _, _ = self.pca_pipe_element.fit(self.X, self.y).transform(self.X)
        X, _, _ = self.pca_pipe_element.inverse_transform(Xt)
        np.testing.assert_array_almost_equal(X, self.X)

    def test_one_hyperparameter_setup(self):
        # sklearn attributes are generated
        self.assertDictEqual(self.pca_pipe_element.hyperparameters, {'PCA__n_components': [1, 2],
                                                                     'PCA__disabled': [False, True]})

        # config_grid is created as expected
        self.assertListEqual(self.pca_pipe_element.generate_config_grid(), [{'PCA__n_components': 1,
                                                                             'PCA__disabled': False},
                                                                            {'PCA__n_components': 2,
                                                                             'PCA__disabled': False},
                                                                            {'PCA__disabled': True}])

    def test_hyperprameter_sanity_check(self):
        with self.assertRaises(ValueError):
            error_element = PipelineElement('PCA', hyperparameters={'kernel': ['rbf', 'linear']})

    def test_more_hyperparameters_setup(self):
        # sklearn attributes are generated
        self.assertDictEqual(self.svc_pipe_element.hyperparameters, {'SVC__C': [0.1, 1],
                                                                     'SVC__kernel': ['rbf', 'sigmoid']})

        # config_grid is created as expected
        self.assertListEqual(self.svc_pipe_element.generate_config_grid(), [{'SVC__C': 0.1, 'SVC__kernel': 'rbf'},
                                                                            {'SVC__C': 0.1, 'SVC__kernel': 'sigmoid'},
                                                                            {'SVC__C': 1, 'SVC__kernel': 'rbf'},
                                                                            {'SVC__C': 1, 'SVC__kernel': 'sigmoid'}])

    def test_no_hyperparameters(self):
        pca_sklearn_element = PCA()
        pca_photon_element = PipelineElement('PCA')

        self.assertDictEqual(pca_sklearn_element.__dict__, pca_photon_element.base_element.__dict__)

    def test_set_params(self):
        config = {'n_components': 3, 'disabled': False}
        self.pca_pipe_element.set_params(**config)
        self.assertFalse(self.pca_pipe_element.disabled)
        self.assertEqual(self.pca_pipe_element.base_element.n_components, 3)
        with self.assertRaises(ValueError):
            self.pca_pipe_element.set_params(**{'any_weird_param': 1})

    def test_get_params(self):
        photon_params = PipelineElement('PCA').get_params()
        del photon_params["name"]
        sk_params = PCA().get_params()
        self.assertDictEqual(photon_params, sk_params)
        self.assertIsNone(PipelineElement.create('any_not_existing_object', object(), hyperparameters={}).get_params())

    def test_set_random_state(self):
        # we handle all elements in one method that is inherited so we capture them all in this test
        random_state = 53
        my_branch = Branch("random_state_branch")
        my_branch += PipelineElement("StandardScaler")
        my_switch = Switch("transformer_Switch")
        my_switch += PipelineElement("LassoFeatureSelection")
        my_switch += PipelineElement("PCA")
        my_branch += my_switch
        my_stack = Stack("Estimator_Stack")
        my_stack += PipelineElement("SVR")
        my_stack += PipelineElement("Ridge")
        my_branch += my_stack
        my_branch += PipelineElement("ElasticNet")

        my_branch.random_state = random_state
        self.assertTrue(my_switch.elements[1].random_state == random_state)
        self.assertTrue(my_switch.elements[1].base_element.random_state == random_state)
        self.assertTrue(my_stack.elements[1].random_state == random_state)
        self.assertTrue(my_stack.elements[1].base_element.random_state == random_state)

    def test_adjusted_delegate_call_transformer(self):
        # check standard transformer
        trans = PipelineElement.create('Transformer', base_element=DummyTransformer(), hyperparameters={})
        X, y, kwargs = trans.transform(self.X, self.y, **self.kwargs)
        self.assertTrue(np.array_equal(X, self.Xt))  # only X should be transformed
        self.assertTrue(np.array_equal(y, self.y))
        self.assertDictEqual(kwargs, self.kwargs)

        # check transformer needs y
        trans = PipelineElement.create('NeedsYTransformer', base_element=DummyNeedsYTransformer(), hyperparameters={})
        X, y, kwargs = trans.transform(self.X, self.y, **self.kwargs)
        self.assertTrue(np.array_equal(X, self.Xt))
        self.assertTrue(np.array_equal(y, self.yt))
        self.assertDictEqual(kwargs, self.kwargs)

        trans = PipelineElement.create('NeedsYTransformer', base_element=DummyNeedsYTransformer(), hyperparameters={})
        X, y, kwargs = trans.transform(self.X, self.y)  # this time without any kwargs
        self.assertTrue(np.array_equal(X, self.Xt))
        self.assertTrue(np.array_equal(y, self.yt))
        self.assertDictEqual(kwargs, {})

        # check transformer needs covariates
        trans = PipelineElement.create('NeedsCovariatesTransformer', base_element=DummyNeedsCovariatesTransformer(),
                                       hyperparameters={})
        X, y, kwargs = trans.transform(self.X, **self.kwargs)
        self.assertTrue(np.array_equal(X, self.Xt))
        self.assertTrue(np.array_equal(kwargs['covariates'], self.kwargst['covariates']))
        self.assertEqual(y, None)

        # check transformer needs covariates and needs y
        trans = PipelineElement.create('NeedsCovariatesAndYTransformer', base_element=DummyNeedsCovariatesAndYTransformer(),
                                       hyperparameters={})
        X, y, kwargs = trans.transform(self.X, self.y, **self.kwargs)
        self.assertTrue(np.array_equal(X, self.Xt))
        self.assertTrue(np.array_equal(y, self.yt))
        self.assertTrue(np.array_equal(kwargs['covariates'], self.kwargst['covariates']))

    def test_adjusted_delegate_call_estimator(self):
        # check standard estimator
        est = PipelineElement.create('Estimator', base_element=DummyEstimator(), hyperparameters={})
        y = est.predict(self.X)
        self.assertTrue(np.array_equal(y, self.Xt)) # DummyEstimator returns X as y predictions

        # check estimator needs covariates
        est = PipelineElement.create('Estimator', base_element=DummyNeedsCovariatesEstimator(), hyperparameters={})
        X = est.predict(self.X, **self.kwargs)
        self.assertTrue(np.array_equal(X, self.Xt))  # DummyEstimator returns X as y predictions

    def test_predict_when_no_transform(self):
        # check standard estimator
        est = PipelineElement.create('Estimator', base_element=DummyEstimator(), hyperparameters={})
        X, y, kwargs = est.transform(self.X)
        self.assertTrue(np.array_equal(X, self.Xt))  # DummyEstimator returns X as y predictions
        self.assertEqual(y, None)

        # check estimator needs covariates
        est = PipelineElement.create('Estimator', base_element=DummyNeedsCovariatesEstimator(), hyperparameters={})
        X, y, kwargs = est.transform(self.X, **self.kwargs)
        self.assertTrue(np.array_equal(X, self.Xt))  # DummyEstimator returns X as y predictions
        self.assertTrue(np.array_equal(kwargs['covariates'], self.kwargs['covariates']))
        self.assertEqual(y, None)

    def test_predict_on_transformer(self):
        est = PipelineElement.create('Estimator', base_element=DummyTransformer(), hyperparameters={})
        with self.assertRaises(BaseException):
            est.predict(self.X)

    def test_copy_me(self):
        svc = PipelineElement('SVC', {'C': [0.1, 1], 'kernel': ['rbf', 'sigmoid']})
        svc.set_params(**{'C': 0.1, 'kernel': 'sigmoid'})
        copy = svc.copy_me()

        self.assertEqual(svc.random_state, copy.random_state)
        self.assertNotEqual(copy.base_element, svc.base_element)
        self.assertDictEqual(elements_to_dict(copy), elements_to_dict(svc))
        self.assertEqual(copy.base_element.C, svc.base_element.C)

        # check if copies are still the same, even when making a copy of a fitted PipelineElement
        copy_after_fit = svc.fit(self.X, self.y).copy_me()
        self.assertDictEqual(elements_to_dict(copy), elements_to_dict(copy_after_fit))

        svc = PipelineElement('SVC', {'C': [0.1, 1], 'kernel': ['rbf', 'sigmoid']})
        copy = svc.copy_me()
        self.assertDictEqual(copy.hyperparameters, {'SVC__C': [0.1, 1], 'SVC__kernel': ['rbf', 'sigmoid']})
        copy.base_element.C = 3
        self.assertNotEqual(svc.base_element.C, copy.base_element.C)

        # test custom element
        custom_element = PipelineElement.create('CustomElement', base_element=DummyNeedsCovariatesEstimator(),
                                                hyperparameters={})
        copy = custom_element.copy_me()
        self.assertDictEqual(elements_to_dict(custom_element), elements_to_dict(copy))

        custom_element2 = PipelineElement.create('MyUnDeepcopyableObject', base_element=GridSearchOptimizer(),
                                                 hyperparameters={})
        with self.assertRaises(Exception):
            custom_element2.copy_me()

    def test_estimator_type(self):
        estimator = PipelineElement('SVC')
        self.assertEqual(estimator._estimator_type, 'classifier')

        estimator = PipelineElement('SVR')
        self.assertEqual(estimator._estimator_type, 'regressor')

        estimator = PipelineElement('PCA')
        self.assertEqual(estimator._estimator_type, None)

        estimator = PipelineElement.create('Dummy', DummyEstimatorWrongType(), {})
        with self.assertRaises(NotImplementedError):
            est_type = estimator._estimator_type

        estimator = PipelineElement.create('Dummy', DummyTransformerWithPredict(), {})
        with self.assertRaises(NotImplementedError):
            est_type = estimator._estimator_type

        estimator = PipelineElement.create('Dummy', DummyEstimatorNoPredict(), {})
        with self.assertRaises(NotImplementedError):
            est_type = estimator._estimator_type

    def test_sanity_check_item_for_add(self):
        valid_type = PipelineElement('StandardScaler')
        valid_type2 = CallbackElement('my_callback', None)
        invalid_type = StandardScaler()
        invalid_type2 = Preprocessing()

        PipelineElement.sanity_check_element_type_for_building_photon_pipes(valid_type, PipelineElement)
        PipelineElement.sanity_check_element_type_for_building_photon_pipes(valid_type2, PipelineElement)

        with self.assertRaises(TypeError):
            PipelineElement.sanity_check_element_type_for_building_photon_pipes(invalid_type, PipelineElement)

        with self.assertRaises(TypeError):
            PipelineElement.sanity_check_element_type_for_building_photon_pipes(invalid_type2, PipelineElement)

        classes_to_test = [Stack, Switch, Branch, Preprocessing]
        for photon_class in classes_to_test:
            # we name it SVC so it suits all classes
            if photon_class is Preprocessing:
                instance = photon_class()
            else:
                instance = photon_class('tmp_instance')
            instance.add(valid_type)
            instance.add(valid_type2)
            with self.assertRaises(TypeError):
                instance.add(invalid_type)
            with self.assertRaises(TypeError):
                instance.add(invalid_type2)


class SwitchTests(unittest.TestCase):

    def setUp(self):
        self.X, self.y = load_breast_cancer(return_X_y=True)
        self.svc = PipelineElement('SVC', {'C': [0.1, 1], 'kernel': ['rbf', 'sigmoid']})
        self.tree = PipelineElement('DecisionTreeClassifier', {'min_samples_split': [2, 3, 4]})
        self.gpc = PipelineElement('GaussianProcessClassifier')
        self.pca = PipelineElement('PCA')

        self.estimator_branch = Branch('estimator_branch', [self.tree.copy_me()])
        self.transformer_branch = Branch('transformer_branch', [self.pca.copy_me()])

        self.estimator_switch = Switch('estimator_switch',
                                       [self.svc.copy_me(), self.tree.copy_me(), self.gpc.copy_me()])
        self.estimator_switch_with_branch = Switch('estimator_switch_with_branch',
                                                   [self.tree.copy_me(), self.estimator_branch.copy_me()])
        self.transformer_switch_with_branch = Switch('transformer_switch_with_branch',
                                                     [self.pca.copy_me(), self.transformer_branch.copy_me()])
        self.switch_in_switch = Switch('Switch_in_switch',
                                       [self.transformer_branch.copy_me(),
                                        self.transformer_switch_with_branch.copy_me()])

    def test_init(self):
        self.assertEqual(self.estimator_switch.name, 'estimator_switch')

    def test_hyperparams(self):
        # assert number of different configs to test
        # each config combi for each element: 4 for SVC and 3 for logistic regression = 7
        self.assertEqual(len(self.estimator_switch.pipeline_element_configurations), 3)
        self.assertEqual(len(self.estimator_switch.pipeline_element_configurations[0]), 4)
        self.assertEqual(len(self.estimator_switch.pipeline_element_configurations[1]), 3)

        # hyperparameters
        self.assertDictEqual(self.estimator_switch.hyperparameters,
                             {'estimator_switch__current_element': [(0, 0), (0, 1), (0, 2), (0, 3),
                                                                    (1, 0), (1, 1), (1, 2), (2, 0)]})

        # config grid
        self.assertListEqual(self.estimator_switch.generate_config_grid(),
                             [{'estimator_switch__current_element': (0, 0)},
                              {'estimator_switch__current_element': (0, 1)},
                              {'estimator_switch__current_element': (0, 2)},
                              {'estimator_switch__current_element': (0, 3)},
                              {'estimator_switch__current_element': (1, 0)},
                              {'estimator_switch__current_element': (1, 1)},
                              {'estimator_switch__current_element': (1, 2)},
                              {'estimator_switch__current_element': (2, 0)}])

    def test_set_params(self):

        # test for grid search
        false_config = {'current_element': 1}
        with self.assertRaises(ValueError):
            self.estimator_switch.set_params(**false_config)

        correct_config = {'current_element': (0, 1)}
        self.estimator_switch.set_params(**correct_config)
        self.assertEqual(self.estimator_switch.base_element.base_element.C, 0.1)
        self.assertEqual(self.estimator_switch.base_element.base_element.kernel, 'sigmoid')

        # test for other optimizers
        smac_config = {'SVC__C': 2, 'SVC__kernel': 'rbf'}
        self.estimator_switch.set_params(**smac_config)
        self.assertEqual(self.estimator_switch.base_element.base_element.C, 2)
        self.assertEqual(self.estimator_switch.base_element.base_element.kernel, 'rbf')

    def test_fit(self):
        np.random.seed(42)
        self.estimator_switch.set_params(**{'current_element': (1, 0)})
        self.estimator_switch.fit(self.X, self.y)
        np.random.seed(42)
        self.tree.set_params(**{'min_samples_split': 2})
        self.tree.fit(self.X, self.y)
        np.testing.assert_array_equal(self.tree.base_element.feature_importances_,
                                      self.estimator_switch.base_element.feature_importances_)

    def test_transform(self):
        self.transformer_switch_with_branch.set_params(**{'current_element': (0, 0)})
        self.transformer_switch_with_branch.fit(self.X, self.y)
        self.pca.fit(self.X, self.y)

        switch_Xt, _, _ = self.transformer_switch_with_branch.transform(self.X)
        pca_Xt, _, _ = self.pca.transform(self.X)
        self.assertTrue(np.array_equal(pca_Xt, switch_Xt))

    def test_predict(self):
        self.estimator_switch.set_params(**{'current_element': (1, 0)})
        np.random.seed(42)
        self.estimator_switch.fit(self.X, self.y)
        self.tree.set_params(**{'min_samples_split': 2})
        np.random.seed(42)
        self.tree.fit(self.X, self.y)

        switch_preds = self.estimator_switch.predict(self.X)
        tree_preds = self.tree.predict(self.X)
        self.assertTrue(np.array_equal(switch_preds, tree_preds))

    def test_predict_proba(self):
        gpc = PipelineElement('GaussianProcessClassifier')
        svc = PipelineElement('SVC')
        switch = Switch('EstimatorSwitch', [gpc, svc])
        switch.set_params(**{'current_element': (0, 0)})
        np.random.seed(42)
        switch_probas = switch.fit(self.X, self.y).predict_proba(self.X)
        np.random.seed(42)
        gpr_probas = self.gpc.fit(self.X, self.y).predict_proba(self.X)
        self.assertTrue(np.array_equal(switch_probas, gpr_probas))

    def test_inverse_transform(self):
        self.transformer_switch_with_branch.set_params(**{'current_element': (0, 0)})
        self.transformer_switch_with_branch.fit(self.X, self.y)
        self.pca.fit(self.X, self.y)
        Xt_pca, _, _ = self.pca.transform(self.X)
        Xt_switch, _, _ = self.transformer_switch_with_branch.transform(self.X)
        X_pca, _, _ = self.pca.inverse_transform(Xt_pca)
        X_switch, _, _ = self.transformer_switch_with_branch.inverse_transform(Xt_switch)

        self.assertTrue(np.array_equal(Xt_pca, Xt_switch))
        self.assertTrue(np.array_equal(X_pca, X_switch))
        np.testing.assert_almost_equal(X_switch, self.X)

    def test_base_element(self):
        switch = Switch('switch', [self.svc, self.tree])
        switch.set_params(**{'current_element': (1, 1)})
        self.assertIs(switch.base_element, self.tree)
        self.assertIs(switch.base_element.base_element, self.tree.base_element)

        # other optimizer
        switch.set_params(**{'DecisionTreeClassifier__min_samples_split': 2})
        self.assertIs(switch.base_element, self.tree)
        self.assertIs(switch.base_element.base_element, self.tree.base_element)

    def test_copy_me(self):
        switches = [self.estimator_switch, self.estimator_switch_with_branch, self.transformer_switch_with_branch,
                    self.switch_in_switch]

        for switch in switches:
            copy = switch.copy_me()

            self.assertEqual(switch.random_state, copy.random_state)

            for i, element in enumerate(copy.elements):
                self.assertNotEqual(copy.elements[i], switch.elements[i])

            switch = elements_to_dict(switch)
            copy = elements_to_dict(copy)

            self.assertDictEqual(copy, switch)

    def test_estimator_type(self):
        pca = PipelineElement('PCA')
        ica = PipelineElement('FastICA')
        svc = PipelineElement('SVC')
        svr = PipelineElement('SVR')
        tree_class = PipelineElement('DecisionTreeClassifier')
        tree_reg = PipelineElement('DecisionTreeRegressor')

        switch = Switch('MySwitch', [pca, svr])
        with self.assertRaises(NotImplementedError):
            est_type = switch._estimator_type

        switch = Switch('MySwitch', [svc, svr])
        with self.assertRaises(NotImplementedError):
            est_type = switch._estimator_type

        switch = Switch('MySwitch', [pca, ica])
        self.assertEqual(switch._estimator_type, None)

        switch = Switch('MySwitch', [tree_class, svc])
        self.assertEqual(switch._estimator_type, 'classifier')

        switch = Switch('MySwitch', [tree_reg, svr])
        self.assertEqual(switch._estimator_type, 'regressor')

        self.assertEqual(self.estimator_switch._estimator_type, 'classifier')
        self.assertEqual(self.estimator_switch_with_branch._estimator_type, 'classifier')
        self.assertEqual(self.transformer_switch_with_branch._estimator_type, None)
        self.assertEqual(self.switch_in_switch._estimator_type, None)

    def test_add(self):
        self.assertEqual(len(self.estimator_switch.elements), 3)
        self.assertEqual(len(self.switch_in_switch.elements), 2)
        self.assertEqual(len(self.transformer_switch_with_branch.elements), 2)

        self.assertEqual(list(self.estimator_switch.elements_dict.keys()), ['SVC', 'DecisionTreeClassifier',
                                                                            'GaussianProcessClassifier'])
        self.assertEqual(list(self.switch_in_switch.elements_dict.keys()), ['transformer_branch',
                                                                            'transformer_switch_with_branch'])

        switch = Switch('MySwitch', [PipelineElement('PCA'), PipelineElement('FastICA')])
        switch = Switch('MySwitch2')
        switch += PipelineElement('PCA')
        switch += PipelineElement('FastICA')

        # test doubled names
        with self.assertRaises(ValueError):
            self.estimator_switch += self.estimator_switch.elements[0]
        self.estimator_switch += PipelineElement("SVC")
        self.assertEqual(self.estimator_switch.elements[-1].name, "SVC2")
        self.estimator_switch += PipelineElement("SVC", hyperparameters={'kernel': ['polynomial', 'sigmoid']})
        self.assertEqual(self.estimator_switch.elements[-1].name, "SVC3")
        self.estimator_switch += PipelineElement("SVR")
        self.assertEqual(self.estimator_switch.elements[-1].name, "SVR")
        self.estimator_switch += PipelineElement("SVC")
        self.assertEqual(self.estimator_switch.elements[-1].name, "SVC4")

        # check that hyperparameters are renamed respectively
        self.assertEqual(self.estimator_switch.pipeline_element_configurations[4][0]["SVC3__kernel"], 'polynomial')

    def test_feature_importances(self):

        self.estimator_switch.set_params(**{'current_element': (1, 0)})
        self.estimator_switch.fit(self.X, self.y)
        self.assertTrue(len(self.estimator_switch.feature_importances_) == self.X.shape[1])

        self.estimator_switch_with_branch.set_params(**{'current_element': (1, 0)})
        self.estimator_switch_with_branch.fit(self.X, self.y)
        self.assertTrue(len(self.estimator_switch_with_branch.feature_importances_) == self.X.shape[1])

        self.estimator_switch.set_params(**{'current_element': (2, 0)})
        self.estimator_switch.fit(self.X, self.y)
        self.assertIsNone(self.estimator_branch.feature_importances_)

        self.switch_in_switch.set_params(**{'current_element': (1, 0)})
        self.switch_in_switch.fit(self.X, self.y)
        self.assertIsNone(self.switch_in_switch.feature_importances_)
        self.estimator_switch.set_params(**{'current_element': (1, 0)})
        self.switch_in_switch.fit(self.X, self.y)
        self.assertIsNone(self.switch_in_switch.feature_importances_)


class BranchTests(unittest.TestCase):

    def setUp(self):
        self.X, self.y = load_breast_cancer(return_X_y=True)
        self.scaler = PipelineElement("StandardScaler", {'with_mean': True})
        self.pca = PipelineElement('PCA', {'n_components': [1, 2]}, test_disabled=True, random_state=3)
        self.tree = PipelineElement('DecisionTreeClassifier', {'min_samples_split': [2, 3, 4]}, random_state=3)

        self.transformer_branch = Branch('MyBranch', [self.scaler, self.pca])
        self.transformer_branch_sklearn = SKPipeline([("SS", StandardScaler()),
                                                      ("PCA", PCA(random_state=3))])
        self.estimator_branch = Branch('MyBranch', [self.scaler, self.pca, self.tree])
        self.estimator_branch_sklearn = SKPipeline([("SS", StandardScaler()),
                                                    ("PCA", PCA(random_state=3)),
                                                    ("Tree", DecisionTreeClassifier(random_state=3))])

    def test_fit(self):
        self.estimator_branch_sklearn.fit(self.X, self.y)
        sk_pred = self.estimator_branch_sklearn.predict(self.X)

        self.estimator_branch.fit(self.X, self.y)
        branch_pred = self.estimator_branch.predict(self.X)

        self.assertTrue(np.array_equal(sk_pred, branch_pred))

    def test_transform(self):
        Xt, _, _ = self.transformer_branch.fit(self.X, self.y).transform(self.X)
        Xt_sklearn = self.transformer_branch_sklearn.fit(self.X, self.y).transform(self.X)
        self.assertTrue(np.array_equal(Xt, Xt_sklearn))

    def test_predict(self):
        y_pred = self.estimator_branch.fit(self.X, self.y).predict(self.X)
        y_pred_sklearn = self.estimator_branch_sklearn.fit(self.X, self.y).predict(self.X)
        np.testing.assert_array_equal(y_pred, y_pred_sklearn)

    def test_predict_proba(self):
        proba = self.estimator_branch.fit(self.X, self.y).predict_proba(self.X)
        proba_sklearn = self.estimator_branch_sklearn.fit(self.X, self.y).predict_proba(self.X)
        np.testing.assert_array_equal(proba, proba_sklearn)

    def test_inverse_transform(self):
        self.estimator_branch.fit(self.X, self.y)
        feature_importances = self.estimator_branch.elements[-1].base_element.feature_importances_
        Xt, _, _ = self.estimator_branch.inverse_transform(feature_importances)
        self.assertEqual(self.X.shape[1], Xt.shape[0])

    def test_no_y_transformers(self):
        stacking_element = Stack("forbidden_stack")
        my_dummy = PipelineElement.create("dummy", DummyNeedsCovariatesAndYTransformer(), {})

        with self.assertRaises(NotImplementedError):
            stacking_element += my_dummy

    def test_copy_me(self):
        branch = Branch('MyBranch', [self.scaler, self.pca])

        copy = branch.copy_me()
        self.assertEqual(branch.random_state, copy.random_state)
        self.assertDictEqual(elements_to_dict(copy), elements_to_dict(branch))

        copy = branch.copy_me()
        copy.elements[1].base_element.n_components = 3
        self.assertNotEqual(copy.elements[1].base_element.n_components, branch.elements[1].base_element.n_components)

        fake_copy = branch
        fake_copy.elements[1].base_element.n_components = 3
        self.assertEqual(fake_copy.elements[1].base_element.n_components, branch.elements[1].base_element.n_components)

    def test_prepare_photon_pipeline(self):
        test_branch = Branch('my_test_branch')
        test_branch += PipelineElement('SimpleImputer')
        test_branch += Switch('my_crazy_switch_bitch')
        test_branch += Stack('my_stacking_stack')
        test_branch += PipelineElement('SVC')

        generated_pipe = test_branch.prepare_photon_pipe(test_branch.elements)

        self.assertEqual(len(generated_pipe.named_steps), 4)
        for idx, element in enumerate(test_branch.elements):
            self.assertIs(generated_pipe.named_steps[element.name], element)
            self.assertIs(generated_pipe.elements[idx][1], test_branch.elements[idx])

    def test_sanity_check_pipe(self):
        test_branch = Branch('my_test_branch')

        def callback_func(X, y, **kwargs):
            pass

        with warnings.catch_warnings(record=True) as w:
            my_callback = CallbackElement('final_element_callback', delegate_function=callback_func)
            test_branch += my_callback
            no_callback_pipe = test_branch.prepare_photon_pipe(test_branch.elements)
            self.assertTrue(no_callback_pipe.elements[-1][1] is my_callback)
            test_branch.sanity_check_pipeline(no_callback_pipe)
            self.assertFalse(no_callback_pipe.elements)
            assert any("Last element of pipeline cannot be callback" in s for s in [e.message.args[0] for e in w])

    def test_prepare_pipeline(self):
        self.assertEqual(len(self.transformer_branch.elements), 2)
        config_grid = {'MyBranch__PCA__n_components': [1, 2],
                       'MyBranch__PCA__disabled': [False, True],
                       'MyBranch__StandardScaler__with_mean': True}
        self.assertDictEqual(config_grid, self.transformer_branch._hyperparameters)

    def test_set_params(self):
        config = {'PCA__n_components': 2,
                  'PCA__disabled': True,
                  'StandardScaler__with_mean': True}
        self.transformer_branch.set_params(**config)
        self.assertTrue(self.transformer_branch.base_element.elements[1][1].disabled)
        self.assertEqual(self.transformer_branch.base_element.elements[1][1].base_element.n_components, 2)
        self.assertEqual(self.transformer_branch.base_element.elements[0][1].base_element.with_mean, True)

        with self.assertRaises(ValueError):
            self.transformer_branch.set_params(**{'any_weird_param': 1})

    def test_estimator_type(self):
        def callback(X, y=None):
            pass

        transformer_branch = Branch('TransBranch', [PipelineElement('PCA'), PipelineElement('FastICA')])
        classifier_branch = Branch('ClassBranch', [PipelineElement('SVC')])
        regressor_branch = Branch('RegBranch', [PipelineElement('SVR')])
        callback_branch = Branch('CallBranch', [PipelineElement('SVR'), CallbackElement('callback', callback)])

        self.assertEqual(transformer_branch._estimator_type, None)
        self.assertEqual(classifier_branch._estimator_type, 'classifier')
        self.assertEqual(regressor_branch._estimator_type, 'regressor')
        self.assertEqual(callback_branch._estimator_type, None)

    def test_add(self):
        branch = Branch('MyBranch', [PipelineElement('PCA', {'n_components': [5]}), PipelineElement('FastICA')])
        self.assertEqual(len(branch.elements), 2)
        self.assertDictEqual(branch._hyperparameters, {'MyBranch__PCA__n_components': [5]})
        branch = Branch('MyBranch')
        branch += PipelineElement('PCA', {'n_components': [5]})
        branch += PipelineElement('FastICA')
        self.assertEqual(len(branch.elements), 2)
        self.assertDictEqual(branch._hyperparameters, {'MyBranch__PCA__n_components': [5]})

        # add doubled item
        branch += PipelineElement('PCA', {'n_components': [10, 20]})
        self.assertEqual(branch.elements[-1].name, 'PCA2')
        self.assertDictEqual(branch.hyperparameters, {'MyBranch__PCA__n_components': [5], 'MyBranch__PCA2__n_components': [10, 20]})

    def test_feature_importances(self):

        self.estimator_branch.fit(self.X, self.y)
        self.assertTrue(len(self.estimator_branch.feature_importances_) == self.X.shape[1])

        self.estimator_branch.elements[-1] = PipelineElement("GaussianProcessClassifier")
        self.estimator_branch.fit(self.X, self.y)
        self.assertIsNone(self.estimator_branch.feature_importances_)

        self.transformer_branch.fit(self.X, self.y)
        self.assertIsNone(self.transformer_branch.feature_importances_)


class StackTests(unittest.TestCase):

    def setUp(self):
        self.X, self.y = load_breast_cancer(return_X_y=True)

        self.pca = PipelineElement('PCA', {'n_components': [5, 10]})
        self.scaler = PipelineElement('StandardScaler', {'with_mean': [True]})
        self.svc = PipelineElement('SVC', {'C': [1, 2]})
        self.tree = PipelineElement('DecisionTreeClassifier', {'min_samples_leaf': [3, 5]})

        self.transformer_branch_1 = Branch('TransBranch1', [self.pca.copy_me()])
        self.transformer_branch_2 = Branch('TransBranch2', [self.scaler.copy_me()])

        self.estimator_branch_1 = Branch('EstBranch1', [self.svc.copy_me()])
        self.estimator_branch_2 = Branch('EstBranch2', [self.tree.copy_me()])

        self.transformer_stack = Stack('TransformerStack', [self.pca.copy_me(), self.scaler.copy_me()])
        self.estimator_stack = Stack('EstimatorStack', [self.svc.copy_me(), self.tree.copy_me()])
        self.transformer_branch_stack = Stack('TransBranchStack', [self.transformer_branch_1.copy_me(),
                                                                   self.transformer_branch_2.copy_me()])
        self.estimator_branch_stack = Stack('EstBranchStack', [self.estimator_branch_1.copy_me(),
                                                               self.estimator_branch_2.copy_me()])

        self.stacks = [([self.pca, self.scaler], self.transformer_stack),
                       ([self.svc, self.tree], self.estimator_stack),
                       ([self.transformer_branch_1, self.transformer_branch_2], self.transformer_branch_stack),
                       ([self.estimator_branch_1, self.estimator_branch_2], self.estimator_branch_stack)]

    def test_copy_me(self):
        for stack in self.stacks:
            stack = stack[1]
            copy = stack.copy_me()
            self.assertEqual(stack.random_state, copy.random_state)
            self.assertFalse(stack.elements[0].__dict__ == copy.elements[0].__dict__)
            self.assertDictEqual(elements_to_dict(stack), elements_to_dict(copy))

    def test_horizontal_stacking(self):
        for stack in self.stacks:
            element_1 = stack[0][0]
            element_2 = stack[0][1]
            stack = stack[1]

            # fit elements
            Xt_1 = element_1.fit(self.X, self.y).transform(self.X, self.y)
            Xt_2 = element_2.fit(self.X, self.y).transform(self.X, self.y)

            Xt = stack.fit(self.X, self.y).transform(self.X, self.y)

            # output of transform() changes depending on whether it is an estimator stack or a transformer stack
            if isinstance(Xt, tuple):
                Xt = Xt[0]
                Xt_1 = Xt_1[0]
                Xt_2 = Xt_2[0]

            if len(Xt_1.shape) == 1:
                Xt_1 = np.reshape(Xt_1, (-1, 1))
                Xt_2 = np.reshape(Xt_2, (-1, 1))

            self.assertEqual(Xt.shape[1], Xt_1.shape[-1] + Xt_2.shape[-1])

    def recursive_assertion(self, element_a, element_b):
        for key in element_a.keys():
            if isinstance(element_a[key], np.ndarray):
                np.testing.assert_array_equal(element_a[key], element_b[key])
            elif isinstance(element_a[key], dict):
                self.recursive_assertion(element_a[key], element_b[key])
            else:
                self.assertEqual(element_a[key], element_b[key])

    def test_fit(self):
        for elements, stack in [([self.pca, self.scaler], self.transformer_stack),
                                ([self.svc, self.tree], self.estimator_stack)]:
            np.random.seed(42)
            stack = stack.fit(self.X, self.y)
            np.random.seed(42)
            for i, element in enumerate(elements):
                element = element.fit(self.X, self.y)
                element_dict = elements_to_dict(element)
                stack_dict = elements_to_dict(stack.elements[i])
                self.recursive_assertion(element_dict, stack_dict)

    def test_transform(self):
        for elements, stack in self.stacks:
            np.random.seed(42)
            Xt_stack, _, _ = stack.fit(self.X, self.y).transform(self.X)
            np.random.seed(42)
            Xt_elements = None
            for i, element in enumerate(elements):
                Xt_element, _, _ = element.fit(self.X, self.y).transform(self.X)
                Xt_elements = PhotonDataHelper.stack_data_horizontally(Xt_elements, Xt_element)
            np.testing.assert_array_equal(Xt_stack, Xt_elements)

    def test_predict(self):
        for elements, stack in [([self.svc, self.tree], self.estimator_stack),
                                ([self.estimator_branch_1, self.estimator_branch_2], self.estimator_branch_stack)]:
            np.random.seed(42)
            stack = stack.fit(self.X, self.y)
            yt_stack = stack.predict(self.X)
            np.random.seed(42)
            Xt_elements = None
            for i, element in enumerate(elements):
                Xt_element = element.fit(self.X, self.y).predict(self.X)
                Xt_elements = PhotonDataHelper.stack_data_horizontally(Xt_elements, Xt_element)
            np.testing.assert_array_equal(yt_stack, Xt_elements)

    def test_predict_proba(self):
        for elements, stack in [([self.svc, self.tree], self.estimator_stack),
                                ([self.estimator_branch_1, self.estimator_branch_2], self.estimator_branch_stack)]:
            np.random.seed(42)
            stack = stack.fit(self.X, self.y)
            yt_stack = stack.predict_proba(self.X)
            np.random.seed(42)
            Xt_elements = None
            for i, element in enumerate(elements):
                Xt_element = element.fit(self.X, self.y).predict_proba(self.X)
                if Xt_element is None:
                    Xt_element = element.fit(self.X, self.y).predict(self.X)
                Xt_elements = PhotonDataHelper.stack_data_horizontally(Xt_elements, Xt_element)
            np.testing.assert_array_equal(yt_stack, Xt_elements)

    def test_inverse_transform(self):
        with self.assertRaises(NotImplementedError):
            self.stacks[0][1].fit(self.X, self.y).inverse_transform(self.X)

    def test_set_params(self):
        trans_config = {'PCA__n_components': 2,
                        'PCA__disabled': True,
                        'StandardScaler__with_mean': True}
        est_config = {'SVC__C': 3,
                      'DecisionTreeClassifier__min_samples_leaf': 1}

        # transformer stack
        self.transformer_stack.set_params(**trans_config)
        self.assertEqual(self.transformer_stack.elements[0].base_element.n_components, 2)
        self.assertEqual(self.transformer_stack.elements[0].disabled, True)
        self.assertEqual(self.transformer_stack.elements[1].base_element.with_mean, True)

        # estimator stack
        self.estimator_stack.set_params(**est_config)
        self.assertEqual(self.estimator_stack.elements[0].base_element.C, 3)
        self.assertEqual(self.estimator_stack.elements[1].base_element.min_samples_leaf, 1)

        with self.assertRaises(ValueError):
            self.estimator_stack.set_params(**{'any_weird_param': 1})

        with self.assertRaises(ValueError):
            self.transformer_stack.set_params(**{'any_weird_param': 1})

    def test_add(self):
        stack = Stack('MyStack', [PipelineElement('PCA', {'n_components': [5]}), PipelineElement('FastICA')])
        self.assertEqual(len(stack.elements), 2)
        self.assertDictEqual(stack._hyperparameters, {'MyStack__PCA__n_components': [5]})
        stack = Stack('MyStack')
        stack += PipelineElement('PCA', {'n_components': [5]})
        stack += PipelineElement('FastICA')
        self.assertEqual(len(stack.elements), 2)
        self.assertDictEqual(stack._hyperparameters, {'MyStack__PCA__n_components': [5]})

        def callback(X, y=None):
            pass

        stack = Stack('MyStack', [PipelineElement('PCA'),
                                  CallbackElement('MyCallback', callback),
                                  Switch('MySwitch', [PipelineElement('PCA'), PipelineElement('FastICA')]),
                                  Branch('MyBranch', [PipelineElement('PCA')])])
        self.assertEqual(len(stack.elements), 4)

        # test doubled item
        with self.assertRaises(ValueError):
            stack += stack.elements[0]

        stack += PipelineElement('PCA', {'n_components': [10, 20]})
        self.assertEqual(stack.elements[-1].name, 'PCA2')
        self.assertDictEqual(stack.hyperparameters, {'MyStack__MySwitch__current_element': [(0, 0), (1, 0)],
                                                     'MyStack__PCA2__n_components': [10, 20]})

    def test_feature_importances(self):
        # single item
        self.estimator_stack.fit(self.X, self.y)
        self.assertIsNone(self.estimator_stack.feature_importances_)

        self.estimator_branch_stack.fit(self.X, self.y)
        self.assertIsNone(self.estimator_branch_stack.feature_importances_)

    def test_use_probabilities(self):
        self.estimator_stack.use_probabilities = True
        self.estimator_stack.fit(self.X, self.y)
        probas = self.estimator_stack.predict(self.X)
        self.assertEqual(probas.shape[1], 3)

        self.estimator_stack.use_probabilities = False
        self.estimator_stack.fit(self.X, self.y)
        preds = self.estimator_stack.predict(self.X)
        self.assertEqual(preds.shape[1], 2)
        probas = self.estimator_stack.predict_proba(self.X)
        self.assertEqual(probas.shape[1], 3)


class PreprocessingTests(unittest.TestCase):

    def test_hyperparameter_add(self):
        pe = Preprocessing()
        with self.assertRaises(ValueError):
            pe.add(PipelineElement('PCA', hyperparameters={'n_components': [1, 5]}))

    def test_predict_warning(self):
        pe = Preprocessing()
        pe.add(PipelineElement('SVC'))
        with warnings.catch_warnings(record=True) as w:
            pe.predict([0, 1, 2])
            assert any("There is no predict function" in s for s in [e.message.args[0] for e in w])


class DataFilterTests(unittest.TestCase):

    def setUp(self):
        self.X, self.y = load_breast_cancer(return_X_y=True)
        self.filter_1 = DataFilter(indices=[0, 1, 2, 3, 4])
        self.filter_2 = DataFilter(indices=[5, 6, 7, 8, 9])

    def test_filter(self):
        Xt_1, y_1, _ = self.filter_1.transform(self.X, self.y)
        Xt_2, y_2, _ = self.filter_2.transform(self.X, self.y)

        self.assertTrue(np.array_equal(self.y, y_1))
        self.assertTrue(np.array_equal(self.y, y_2))
        self.assertTrue(np.array_equal(Xt_1, self.X[:, :5]))
        self.assertTrue(np.array_equal(Xt_2, self.X[:, 5:10]))


class CallbackElementTests(unittest.TestCase):

    def setUp(self):
        def callback(X, y=None, **kwargs):
            self.assertEqual(X.shape, (569, 30))
            print("Shape of transformed data: {}".format(X.shape))

        def predict_callback(X, y=None, **kwargs):
            self.assertEqual(X.shape, (569, ))
            print('Shape of predictions: {}'.format(X.shape))

        def callback_test_equality(X, y=None, **kwargs):
            self.assertTrue(np.array_equal(self.X, X))
            if y is not None:
                self.assertListEqual(self.y.tolist(), y.tolist())

        self.X, self.y = load_breast_cancer(return_X_y=True)

        self.clean_pipeline = PhotonPipeline(elements=[('PCA', PipelineElement('PCA')),
                                                       ('LogisticRegression', PipelineElement('LogisticRegression'))])
        self.callback_pipeline = PhotonPipeline(elements=[('First', CallbackElement('First', callback)),
                                                          ('PCA', PipelineElement('PCA')),
                                                          ('Second', CallbackElement('Second', callback)),
                                                          ('LogisticRegression', PipelineElement('LogisticRegression'))])
        self.clean_branch_pipeline = PhotonPipeline(elements=[('MyBranch',
                                                               Branch('MyBranch', [PipelineElement('PCA')])),
                                                              ('LogisticRegression',
                                                               PipelineElement('LogisticRegression'))])
        self.callback_branch_pipeline = PhotonPipeline(elements=[('First', CallbackElement('First', callback)),
                                                                 ('MyBranch', Branch('MyBranch', [CallbackElement('Second',
                                                                                                                  callback),
                                                                                                  PipelineElement('PCA')])),
                                                                 ('Fourth', CallbackElement('Fourth', callback)),
                                                                 ('LogisticRegression',
                                                                  PipelineElement('LogisticRegression'))])
        self.callback_branch_pipeline_error = PhotonPipeline(elements=[('First', CallbackElement('First', callback)),
                                                                       ('MyBranch', Branch('MyBranch', [CallbackElement('Second',
                                                                                                                        callback),
                                                                                                        PipelineElement('PCA'),
                                                                                                        CallbackElement('Third',
                                                                                                                        callback)])),
                                                                       ('Fourth', CallbackElement('Fourth', callback)),
                                                                       ('LogisticRegression',
                                                                        PipelineElement('LogisticRegression')),
                                                                       ('Fifth', CallbackElement('Fifth', predict_callback))])
        # test that data is unaffected from pipeline
        self.callback_after_callback_pipeline = PhotonPipeline([('Callback1', CallbackElement('Callback1', callback)),
                                                                ('Callback2', CallbackElement('Callback2', callback_test_equality)),
                                                                ('StandarcScaler', PipelineElement('StandardScaler'),
                                                                 ('SVR', PipelineElement('SVR')))])

    def test_callback(self):
        pipelines = [self.clean_pipeline, self.callback_pipeline, self.clean_branch_pipeline,
                     self.callback_branch_pipeline, self.callback_after_callback_pipeline]

        for pipeline in pipelines:
            pipeline.fit(self.X, self.y).predict(self.X)

        with warnings.catch_warnings(record=True) as w:
            self.callback_branch_pipeline_error.fit(self.X, self.y).predict(self.X)
            assert any("Last element of pipeline cannot be callback" in s for s in [e.message.args[0] for e in w])




from sklearn.datasets import load_breast_cancer
from sklearn.model_selection import KFold

from photonai.base import Hyperpipe, PipelineElement
from photonai.optimization import FloatRange, Categorical, IntegerRange

my_pipe = Hyperpipe('basic_svm_pipe',
                    inner_cv=KFold(n_splits=5),
                    outer_cv=KFold(n_splits=3),
                    optimizer='sk_opt',
                    optimizer_params={'n_configurations': 15},
                    metrics=['accuracy', 'precision', 'recall', 'balanced_accuracy'],
                    best_config_metric='accuracy',
                    project_folder='./tmp')

my_pipe.add(PipelineElement('StandardScaler'))

my_pipe += PipelineElement('PCA',
                           hyperparameters={'n_components': IntegerRange(10, 30)},
                           test_disabled=True)

my_pipe += PipelineElement('SVC',
                           hyperparameters={'kernel': Categorical(['rbf', 'linear']),
                                            'C': FloatRange(1, 6)},
                           gamma='scale')

X, y = load_breast_cancer(return_X_y=True)
my_pipe.fit(X, y)



import warnings
from sklearn.model_selection import StratifiedKFold, StratifiedShuffleSplit
from sklearn.exceptions import UndefinedMetricWarning
from imblearn.datasets import fetch_datasets

from photonai.base import Hyperpipe, PipelineElement
from photonai.optimization import Categorical

# Since we test very imbalanced data, we want to ignore some metric based zero-divisions.
warnings.filterwarnings("ignore", category=UndefinedMetricWarning)

# example of imbalanced dataset
dataset = fetch_datasets()['coil_2000']
X, y = dataset.data, dataset.target
# ratio class 0: 6%, class 1: 94%

my_pipe = Hyperpipe('balancing_pipe',
                    optimizer='grid_search',
                    metrics=['accuracy', 'precision', 'recall', 'f1_score'],
                    best_config_metric='f1_score',
                    outer_cv=StratifiedKFold(n_splits=3),
                    inner_cv=StratifiedShuffleSplit(n_splits=5, test_size=0.2),
                    verbosity=1,
                    project_folder='./tmp/')


# ADD ELEMENTS TO YOUR PIPELINE
my_pipe += PipelineElement('StandardScaler')

tested_methods = Categorical(['RandomOverSampler', 'SMOTEENN', 'SVMSMOTE',
                              'BorderlineSMOTE', 'SMOTE', 'ClusterCentroids'])
my_pipe += PipelineElement('ImbalancedDataTransformer',
                           hyperparameters={'method_name': tested_methods},
                           test_disabled=True)

my_pipe += PipelineElement("RandomForestClassifier", n_estimators=200)

# NOW TRAIN YOUR PIPELINE
my_pipe.fit(X, y)

# Possible values for method_name:
# imbalance_type = OVERSAMPLING:
#        - ADASYN
#        - BorderlineSMOTE
#        - KMeansSMOTE
#        - RandomOverSampler
#        - SMOTE
#        - SMOTENC
#        - SVMSMOTE
#
# imbalance_type = UNDERSAMPLING:
#        - ClusterCentroids,
#        - RandomUnderSampler,
#        - NearMiss,
#        - InstanceHardnessThreshold,
#        - CondensedNearestNeighbour,
#        - EditedNearestNeighbours,
#        - RepeatedEditedNearestNeighbours,
#        - AllKNN,
#        - NeighbourhoodCleaningRule,
#        - OneSidedSelection
#
# imbalance_type = COMBINE:
#        - SMOTEENN,
#        - SMOTETomek



import numpy as np
from sklearn.metrics import f1_score
from sklearn.datasets import fetch_openml
from sklearn.model_selection import KFold
from keras.metrics import Accuracy

from photonai.base import Hyperpipe, PipelineElement


# you can have a simple delegate
def custom_metric(y_true, y_pred):
    def hot_encoding(targets, nclasses):
        """Convert indices to one-hot encoded labels."""
        targets = np.array(targets).reshape(-1)
        return np.eye(nclasses)[targets]
    return f1_score(hot_encoding(y_true, 3), hot_encoding(y_pred, 3), average='macro')


my_pipe = Hyperpipe('custom_metric_project',
                    optimizer='sk_opt',
                    optimizer_params={'n_configurations': 25},
                    # and here is how to register it in photonai
                    metrics=[('custom_metric', custom_metric), Accuracy, 'accuracy'],
                    best_config_metric='custom_metric',
                    outer_cv=KFold(n_splits=3),
                    inner_cv=KFold(n_splits=3),
                    allow_multidim_targets=True,
                    project_folder='./tmp/')

my_pipe += PipelineElement('StandardScaler')
my_pipe += PipelineElement('SVC', kernel='rbf')

X, y = fetch_openml("cars1", return_X_y=True)
my_pipe.fit(X.values, y.values.astype(int))



import numpy as np

from photonai.modelwrapper.imbalanced_data_transformer import ImbalancedDataTransformer
from test.modelwrapper_tests.test_base_model_wrapper import BaseModelWrapperTest

from imblearn.over_sampling.tests import test_smote
from imblearn.combine.tests import test_smote_tomek
from imblearn.under_sampling._prototype_selection.tests import test_instance_hardness_threshold


class ImbalancedDataTransformTest(BaseModelWrapperTest):
    """
    Tests based on implemented version by imblearn.
    For original implementation see:
    https://github.com/scikit-learn-contrib/imbalanced-learn/blob/master/imblearn/over_sampling/tests/
    """

    def setUp(self):
        super(ImbalancedDataTransformTest, self).setUp()
        self.model_wrapper = ImbalancedDataTransformer()

    def test_strategy(self):
        with self.assertRaises(ValueError):
            ImbalancedDataTransformer(method_name="something")


    def test_strategy_oversampling(self):
        """
        sample test of different functions based on imblearn implementation for oversampling methods.
        """
        sampling_strategy = {0: 9, 1: 12}
        imbalanced_data_transformer = ImbalancedDataTransformer(method_name='SMOTE',
                                                                sampling_strategy = {0: 9, 1: 12},
                                                                random_state = test_smote.RND_SEED)

        # test_sample_regular_half() -> smote
        X_resampled, y_resampled = imbalanced_data_transformer.fit_transform(test_smote.X, test_smote.Y)
        X_gt = np.array([[0.11622591, -0.0317206], [0.77481731, 0.60935141], [
            1.25192108, -0.22367336
        ], [0.53366841, -0.30312976], [1.52091956, -0.49283504], [
                             -0.28162401, -2.10400981
                         ], [0.83680821, 1.72827342], [0.3084254, 0.33299982], [
                             0.70472253, -0.73309052
                         ], [0.28893132, -0.38761769], [1.15514042, 0.0129463], [
                             0.88407872, 0.35454207
                         ], [1.31301027, -0.92648734], [-1.11515198, -0.93689695], [
                             -0.18410027, -0.45194484
                         ], [0.9281014, 0.53085498], [-0.14374509, 0.27370049],
                         [-0.41635887, -0.38299653], [0.08711622, 0.93259929],
                         [1.70580611, -0.11219234], [0.36784496, -0.1953161]])
        y_gt = np.array(
            [0, 1, 0, 0, 0, 1, 1, 1, 1, 1, 1, 0, 0, 1, 1, 1, 1, 0, 1, 0, 0])
        test_smote.assert_allclose(X_resampled, X_gt, rtol=test_smote.R_TOL)
        test_smote.assert_array_equal(y_resampled, y_gt)

    def test_strategy_undersampling(self):
        """
        sample test of different functions based on imblearn implementation for undersampling methods.
        """
        imbalanced_data_transformer = ImbalancedDataTransformer(method_name='InstanceHardnessThreshold',
                                                                estimator=test_instance_hardness_threshold.ESTIMATOR,
                                                                sampling_strategy={0: 6, 1: 8},
                                                                random_state=test_instance_hardness_threshold.RND_SEED)

        X_resampled, y_resampled = imbalanced_data_transformer.fit_resample(test_instance_hardness_threshold.X,
                                                                            test_instance_hardness_threshold.Y)
        assert X_resampled.shape == (15, 2)
        assert y_resampled.shape == (15,)

    def test_strategy_combine(self):
        """
        sample test of different functions based on imblearn implementation for oversampling methods.
        """
        imbalanced_data_transformer = ImbalancedDataTransformer(method_name='SMOTETomek',
                                                              random_state=test_smote_tomek.RND_SEED)
        X_resampled, y_resampled = imbalanced_data_transformer.fit_resample(test_smote_tomek.X, test_smote_tomek.Y)
        X_gt = np.array(
            [
                [0.68481731, 0.51935141],
                [1.34192108, -0.13367336],
                [0.62366841, -0.21312976],
                [1.61091956, -0.40283504],
                [-0.37162401, -2.19400981],
                [0.74680821, 1.63827342],
                [0.61472253, -0.82309052],
                [0.19893132, -0.47761769],
                [1.40301027, -0.83648734],
                [-1.20515198, -1.02689695],
                [-0.23374509, 0.18370049],
                [-0.00288378, 0.84259929],
                [1.79580611, -0.02219234],
                [0.38307743, -0.05670439],
                [0.70319159, -0.02571667],
                [0.75052536, -0.19246518],
            ]
        )
        y_gt = np.array([1, 0, 0, 0, 1, 1, 1, 1, 0, 1, 1, 1, 0, 0, 0, 0])
        test_smote_tomek.assert_allclose(X_resampled, X_gt, rtol=test_smote_tomek.R_TOL)
        test_smote_tomek.assert_array_equal(y_resampled, y_gt)



from sklearn.datasets import load_boston
from sklearn.model_selection import KFold
from photonai.base import Hyperpipe, PipelineElement, CallbackElement


# DEFINE CALLBACK ELEMENT
def my_monitor(X, y=None, **kwargs):
    print(X.shape)

    # here is a useless statement where you can easily set a breakpoint
    # and do fancy developer stuff
    debug = True

my_pipe = Hyperpipe('monitoring_pipe',
                    optimizer='grid_search',
                    metrics=['mean_squared_error', 'pearson_correlation'],
                    best_config_metric='mean_squared_error',
                    outer_cv=KFold(n_splits=3),
                    inner_cv=KFold(n_splits=3),
                    verbosity=1,
                    project_folder='./tmp/')

my_pipe += PipelineElement('StandardScaler')

my_pipe += PipelineElement('SamplePairingClassification',
                           hyperparameters={'draw_limit': [500, 1000, 10000]},
                           distance_metric='euclidean',
                           generator='nearest_pair',
                           test_disabled=True)

# here we inspect the data after augmentation
my_pipe += CallbackElement("monitor", my_monitor)

my_pipe += PipelineElement('RandomForestRegressor', hyperparameters={'n_estimators': [10, 100]})

X, y = load_boston(return_X_y=True)
my_pipe.fit(X, y)



from photonai.base import Hyperpipe
from sklearn.datasets import load_breast_cancer

X, _ = load_breast_cancer(True)

# After optimization is finished, PHOTONAI saves the user's pipeline
# fitted with the best hyperparameter configuration found
# as "photon_best_model.photon" in the project's result folder.
# this is done automatically, however the use may do so manually by calling
# my_pipe.save_optimum_pipe('/home/photon_user/photon_test/optimum_pipe.photon')

my_pipe = Hyperpipe.load_optimum_pipe("full_path/to/photon_best_model.photon")
predictions = my_pipe.predict(X)



from sklearn.neural_network import MLPClassifier

from sklearn.base import BaseEstimator, ClassifierMixin


class PhotonMLPClassifier(BaseEstimator, ClassifierMixin):

    def __init__(self, layer_1: int = 10, layer_2: int = None, layer_3: int = None, layer_4: int = None,
                 activation='relu', solver='adam', alpha=0.0001,
                 batch_size='auto', learning_rate='constant', learning_rate_init=0.001,
                 power_t=0.5, max_iter=200, shuffle=True, random_state=None, tol=0.0001,
                 verbose=False, warm_start=False, momentum=0.9, nesterovs_momentum=True,
                 early_stopping=False, validation_fraction=0.1, beta_1=0.9, beta_2=0.999,
                 epsilon=1e-08, n_iter_no_change=10):
        self.layer_1 = layer_1
        self.layer_2 = layer_2
        self.layer_3 = layer_3
        self.layer_4 = layer_4

        self.activation = activation
        self.solver = solver
        self.alpha = alpha
        self.batch_size = batch_size
        self.learning_rate = learning_rate
        self.learning_rate_init = learning_rate_init
        self.power_t = power_t
        self.max_iter = max_iter
        self.shuffle = shuffle
        self.random_state = random_state
        self.tol = tol
        self.verbose = verbose
        self.warm_start = warm_start
        self.momentum = momentum
        self.nesterovs_momentum = nesterovs_momentum
        self.early_stopping = early_stopping
        self.validation_fraction = validation_fraction
        self.beta_1 = beta_1
        self.beta_2 = beta_2
        self.epsilon = epsilon
        self.n_iter_no_change = n_iter_no_change

        self.model = None

    def fit(self, X, y=None, **kwargs):
        hidden_layer_sizes = list()
        for i, layer in enumerate([self.layer_1, self.layer_2, self.layer_3, self.layer_4]):
            if layer:
                hidden_layer_sizes.append(layer)
            else:
                if i == 0:
                    raise ValueError("Number of neurons in first hidden layer is zero.")
                else:
                    break

        self.model = MLPClassifier(hidden_layer_sizes=hidden_layer_sizes,
                                   activation=self.activation,
                                   solver=self.solver,
                                   alpha=self.alpha,
                                   batch_size=self.batch_size,
                                   learning_rate=self.learning_rate,
                                   learning_rate_init=self.learning_rate_init,
                                   power_t=self.power_t,
                                   max_iter=self.max_iter,
                                   shuffle=self.shuffle,
                                   random_state=self.random_state,
                                   tol=self.tol,
                                   verbose=self.verbose,
                                   warm_start=self.warm_start,
                                   momentum=self.momentum,
                                   nesterovs_momentum=self.nesterovs_momentum,
                                   early_stopping=self.early_stopping,
                                   validation_fraction=self.validation_fraction,
                                   beta_1=self.beta_1,
                                   beta_2=self.beta_2,
                                   epsilon=self.epsilon,
                                   n_iter_no_change=self.n_iter_no_change)
        self.model.fit(X, y)
        return self

    def predict(self, X):
        return self.model.predict(X)

    def predict_proba(self, X):
        return self.model.predict_proba(X)



import numpy as np
from sklearn.datasets import load_breast_cancer
from sklearn.model_selection import GroupKFold, GroupShuffleSplit

from photonai.base import Hyperpipe, PipelineElement
from photonai.optimization import FloatRange, Categorical

# WE USE THE BREAST CANCER SET FROM SKLEARN
X, y = load_breast_cancer(return_X_y=True)

groups = np.random.random_integers(0, 3, (len(y), ))

# DESIGN YOUR PIPELINE
my_pipe = Hyperpipe('group_split_pipe',
                    optimizer='grid_search',
                    metrics=['accuracy', 'precision', 'recall'],
                    best_config_metric='accuracy',
                    outer_cv=GroupKFold(n_splits=4),
                    inner_cv=GroupShuffleSplit(n_splits=10),
                    verbosity=1,
                    project_folder='./tmp/')

# ADD ELEMENTS TO YOUR PIPELINE
# first normalize all features
my_pipe += PipelineElement('StandardScaler')
# then do feature selection using a PCA, specify which values to try in the hyperparameter search
my_pipe += PipelineElement('PCA', hyperparameters={'n_components': [5, 10, None]}, test_disabled=True)
# engage and optimize the good old SVM for Classification
my_pipe += PipelineElement('SVC', hyperparameters={'kernel': Categorical(['rbf', 'linear']),
                                                   'C': FloatRange(0.5, 2, "linspace", num=5)})

# NOW TRAIN YOUR PIPELINE
my_pipe.fit(X, y, groups=groups)



import warnings
import keras
from sklearn.base import BaseEstimator

from photonai.photonlogger.logger import logger


class KerasBaseEstimator(BaseEstimator):
    """
    base class for all Keras wrappers
    """

    def __init__(self,
                 model: keras.Model = None,
                 epochs: int = 10,
                 nn_batch_size: int = 32,
                 callbacks: list = None,
                 validation_split: float = 0.1,
                 verbosity: int = 0):
        self.model = model
        self.epochs = epochs
        self.nn_batch_size = nn_batch_size
        self.verbosity = verbosity
        self.validation_split = validation_split
        if callbacks:
            self.callbacks = callbacks
        else:
            self.callbacks = []

        # in reloading context self.model could be None
        if self.model is not None:
            # saving initial weights in order to be able to reset the weights before each fit
            # with this approach pre-trained weights can be set
            self.init_weights = model.get_weights()

            # if a non keras layer is in model, it needs to be registered with keras
            for layer in self.model.layers:
                if not layer.__module__.startswith('keras'):
                    keras.utils.get_custom_objects()[type(layer).__name__] = layer.__class__

    def fit(self, X, y):
        # set weights to initial weights to achieve a weight reset
        self.model.set_weights(self.init_weights)

        # allow labels to be encoded before being passed to the model
        # by default self.encode_targets returns identity of y
        y = self.encode_targets(y)

        # use validation split only when size of training set is above 100
        if X.shape[0] < 100 and self.validation_split is not None:
            msg = 'Cannot use validation split because of small sample size.'
            logger.warning(msg)
            warnings.warn(msg)
        self.model.fit(X, y,
                       batch_size=self.nn_batch_size,
                       epochs=self.epochs,
                       validation_split=self.validation_split if X.shape[0] > 100 else None,
                       callbacks=self.callbacks,
                       verbose=self.verbosity)
        return self

    def predict_proba(self, X):
        """
        Predict probabilities
        :param X: array-like
        :return: predicted values, array
        """
        return self.model.predict(X, batch_size=self.nn_batch_size)

    def encode_targets(self, y):
        return y

    def save(self, filename):
        # serialize model to JSON
        model_json = self.model.to_json()
        with open(filename + ".json", "w") as json_file:
            json_file.write(model_json)
        # serialize weights to HDF5
        self.model.save_weights(filename + ".h5")

    def load(self, filename):
        # load json and create model
        json_file = open(filename + '.json', 'r')
        loaded_model_json = json_file.read()
        json_file.close()
        loaded_model = keras.models.model_from_json(loaded_model_json)

        # load weights into new model
        loaded_model.load_weights(filename + ".h5")
        self.model = loaded_model
        self.init_weights = self.model.get_weights()



from photonai.base import PipelineElement
from photonai.optimization import SkOptOptimizer, IntegerRange, Categorical, FloatRange, BooleanSwitch
from photonai.optimization.hyperparameters import NumberRange
from ..grid_search_tests.test_grid_search import GridSearchOptimizerTest

import warnings
import time


class SkOptOptimizerTest(GridSearchOptimizerTest):

    def setUp(self):
        self.pipeline_elements = [PipelineElement("StandardScaler"),
                                  PipelineElement('PCA', hyperparameters={'n_components': IntegerRange(5, 20)}),
                                  PipelineElement("SVC", hyperparameters={'C': FloatRange(1, 100)})]
        self.optimizer = SkOptOptimizer()
        self.optimizer_name = "sk_opt"
        self.optimizer_params = {'n_configurations': 10}

    def test_ask_advanced(self):
        with self.assertRaises(ValueError):
            super(SkOptOptimizerTest, self).test_ask_advanced()

    def test_empty_hspace(self):
        with warnings.catch_warnings(record=True) as w:
            self.optimizer.prepare([], True)
            self.assertIsNone(self.optimizer.optimizer)
            assert any("Did not find any" in s for s in [e.message.args[0] for e in w])

    def test_eliminate_one_value_hyperparams(self):
        pipeline_elements = [PipelineElement('PCA', hyperparameters={'n_components': Categorical([5])}),
                             PipelineElement("SVC", hyperparameters={'kernel': ['rbf'],
                                                                     'shrinking': BooleanSwitch(),
                                                                     'C':FloatRange(0.5, 2, range_type='linspace'),
                                                                     'tol': FloatRange(0.1, 1, range_type='logspace')})]
        with warnings.catch_warnings(record=True) as w:
            self.optimizer.prepare(pipeline_elements, True)
            assert any("PHOTONAI has detected some" in s for s in [e.message.args[0] for e in w])
        self.assertIn('SVC__C', self.optimizer.hyperparameter_list)
        self.assertIn('SVC__shrinking', self.optimizer.hyperparameter_list)
        self.assertNotIn('PCA__n_components', self.optimizer.hyperparameter_list)
        self.assertNotIn('SVC__kernel', self.optimizer.hyperparameter_list)

        self.optimizer.prepare([pipeline_elements[0]], True)
        self.assertIsNone(self.optimizer.optimizer)

        i = 0
        for val in self.optimizer.ask:
            self.assertFalse(val)
            i += 1
        self.assertEqual(i, 1)

    def test_geomspace(self):
        pipeline_elements = [PipelineElement("SVC", hyperparameters={'kernel': ['rbf', 'poly'],
                                                                     'C': FloatRange(0.1, 0.5, range_type='geomspace'),
                                                                     'tol': FloatRange(0.001, 0.01)})]
        with self.assertRaises(ValueError):
            self.optimizer.prepare(pipeline_elements, True)

    def test_unsported_param(self):
        pipeline_elements = [PipelineElement("SVC", hyperparameters={'kernel': Categorical(['rbf', 'poly', 'sigmoid']),
                                                                     'C': NumberRange(1, 3, range_type='range')})]
        with self.assertRaises(ValueError):
            self.optimizer.prepare(pipeline_elements, True)

    def test_time_limit(self):
        self.optimizer = SkOptOptimizer(limit_in_minutes=0.05, n_configurations=1000)  # 3 seconds
        self.optimizer.prepare(pipeline_elements=self.pipeline_elements, maximize_metric=True)
        configs = []
        start = time.time()
        for config in self.optimizer.ask:
            configs.append(config)
            time.sleep(0.005)
        stop = time.time()
        self.assertAlmostEqual(stop-start, 3, 0)



import numpy as np
from sklearn.preprocessing import LabelEncoder as SKLabelEncoder


class LabelEncoder(SKLabelEncoder):
    """
    Suitable version of the scikit-learn LabelEncoder for PHOTONAI.
    Since the pipeline process streams the underlying samples to
    every transformer, this class is required.

    """
    def __init__(self):
        """Initialize the object."""
        super(LabelEncoder, self).__init__()
        self.needs_y = True

    def fit(self, X: np.ndarray, y: np.ndarray = None, **kwargs):
        """
        Call of the underlying sklearn.fit(y) method.

        Parameters:
            X:
                The input samples of shape [n_samples, n_features].

            y:
                The input targets of shape [n_samples, 1].

            **kwargs:
                Ignored input.

        """
        super(LabelEncoder, self).fit(y)
        return self

    def transform(self, X: np.ndarray, y: np.ndarray = None, **kwargs) -> (np.ndarray, np.ndarray):
        """
        Call of the underlying sklearn.transform(y) method.

        Parameters:
            X:
                The input samples of shape [n_samples, n_features].

            y:
                The input targets of shape [n_samples, 1].

            **kwargs:
                Ignored input.

        Returns:
            Original X and encoded y.

        """
        yt = super(LabelEncoder, self).transform(y)
        return X, yt

    def fit_transform(self, X: np.ndarray, y: np.ndarray = None, **kwargs) -> (np.ndarray, np.ndarray):
        """
        Call of the underlying sklearn.fit_transform(y) method.

        Parameters:
            X:
                The input samples of shape [n_samples, n_features].

            y:
                The input targets of shape [n_samples, 1].

            **kwargs:
                Ignored input.

        Returns:
            Original X and encoded y.

        """
        return super(LabelEncoder, self).fit_transform(y)



import time

from photonai.base import PipelineElement
from photonai.optimization import RandomSearchOptimizer, IntegerRange
from ..grid_search_tests.test_grid_search import GridSearchOptimizerTest


class RandomSearchOptimizerTest(GridSearchOptimizerTest):

    def setUp(self):
        self.pipeline_elements = [PipelineElement("StandardScaler"),
                                  PipelineElement('PCA', hyperparameters={'n_components': IntegerRange(5, 20)}),
                                  PipelineElement("SVC")]
        self.optimizer = RandomSearchOptimizer(n_configurations=5)
        self.optimizer_name = 'random_search'

    def test_parameter_k(self):
        """Test for parameter n_configuration and k."""
        self.optimizer = RandomSearchOptimizer(n_configurations=3)
        self.optimizer.prepare(pipeline_elements=self.pipeline_elements, maximize_metric=True)
        configs = []
        for config in self.optimizer.ask:
            configs.append(config)
        self.assertEqual(len(configs), 3)
        self.optimizer = RandomSearchOptimizer(n_configurations=500)
        self.optimizer.prepare(pipeline_elements=self.pipeline_elements, maximize_metric=True)
        configs = []
        for config in self.optimizer.ask:
            configs.append(config)
        self.assertEqual(len(configs), 500)

    def test_constraint_obligation(self):
        with self.assertRaises(ValueError):
            RandomSearchOptimizer(n_configurations=-1, limit_in_minutes=-1)

    def test_time_limit(self):
        self.optimizer = RandomSearchOptimizer(limit_in_minutes=0.05)  # 3 seconds
        self.optimizer.prepare(pipeline_elements=self.pipeline_elements, maximize_metric=True)
        configs = []
        start = time.time()
        for config in self.optimizer.ask:
            configs.append(config)
        stop = time.time()
        self.assertAlmostEqual(stop-start, 3, 0)

    def test_run(self):
        pass



"""
PHOTON classes for testing a specific hyperparameter configuration and calculating the performance metrics.
"""

from .results_handler import ResultsHandler


# __all__ = ("InnerFoldManager",
           # "Scorer")



from sklearn.datasets import load_boston
from sklearn.model_selection import KFold

from photonai.base import Hyperpipe, PipelineElement, Switch
from photonai.optimization import IntegerRange

X, y = load_boston(return_X_y=True)

my_pipe = Hyperpipe('feature_selection',
                    optimizer='grid_search',
                    metrics=['mean_squared_error', 'pearson_correlation', 'mean_absolute_error', 'explained_variance'],
                    best_config_metric='mean_squared_error',
                    outer_cv=KFold(n_splits=3),
                    inner_cv=KFold(n_splits=3),
                    verbosity=1,
                    project_folder='./tmp/')

my_pipe += PipelineElement('StandardScaler')

lasso = PipelineElement('LassoFeatureSelection',
                        hyperparameters={'percentile': [0.1, 0.2, 0.3]}, alpha=1)
f_regression = PipelineElement('FRegressionSelectPercentile',
                               hyperparameters={'percentile': [10, 20, 30]})

my_pipe += Switch('FeatureSelection', [lasso, f_regression])
my_pipe += PipelineElement('RandomForestRegressor',
                           hyperparameters={'n_estimators': IntegerRange(10, 50)})
my_pipe.fit(X, y)



import unittest
import numpy as np

from numpy.testing import assert_array_equal
from photonai.modelwrapper.source_splitter import SourceSplitter


class FeatureEncoderTests(unittest.TestCase):

    def setUp(self):
        self.source_splitter = SourceSplitter(column_indices=[1,2])
        self.X = np.array([["a", 3, "b"], ["x", 5, "b"], ["a", 3, "c"]], dtype=object)
        self.values = np.array([1, 2, 1])
        self.X_two_columns = np.array([[1, 2], [2, 3], [3, 4]])

    def test_fit(self):
        NotImplementedError()

    def test_transform(self):
        result = self.source_splitter.transform(self.X)
        assert_array_equal(result, np.array([[3, "b"], [5, "b"], [3, "c"]], dtype=object))

        # selber abfangen?
        with self.assertRaises(IndexError):
            self.source_splitter.transform(self.X_two_columns)

    def test_inverse_transform(self):
        NotImplementedError()

    def test_fit_transform(self):
        result = self.source_splitter.transform(self.X)
        assert_array_equal(result, np.array([[3, "b"], [5, "b"], [3, "c"]], dtype=object))







import datetime
import warnings
import numpy as np
import json

from photonai.helper.helper import PhotonDataHelper, print_double_metrics, print_metrics
from photonai.optimization import DummyPerformanceConstraint
from photonai.photonlogger.logger import logger
from photonai.processing.inner_folds import InnerFoldManager
from photonai.processing.photon_folds import FoldInfo
from photonai.processing.results_structure import MDBInnerFold, MDBScoreInformation
from photonai.optimization.base_optimizer import PhotonSlaveOptimizer, PhotonMasterOptimizer

warnings.filterwarnings('ignore', category=DeprecationWarning)
warnings.filterwarnings('ignore', category=FutureWarning)


class OuterFoldManager:
    """Outer Fold manager.

    Controls the tasks over a specified Outer Fold.
    It is responsible for generating the split in the outer folds
    and triggering the hyperparameter optimization process.
    An Objective Function is provided for this purpose.
    This defines a black box function over the outer_fold data.

    Parameters
    ----------
    pipe: PhotonPipeline
        Defined pipeline structure for optimization.

    optimization_info: Optimization
        Contains the information how the black box function is solved.
        Depending on the algorithms and metrics, the objective function is adapted to it.

    outer_fold_id: UUID
        Unique ID for this object.

    cache_folder, str or None, default=None
        Folder for storing information in multiprocessing case.

    cache_updater, default=None
        The object that takes active access to the cache structure.
        Only in the multiprocess case.

    dummy_estimator: DummyClassifier, DummyRegressor or None, default=None
        To be able to classify the results,
        they are compared against a dummy performance.
        Since there are exceptions to the calculation,
        this does not necessarily have to be passed.

    result_obj: MDBOuterFold, default=None
        Contains the memory structure for this object.
        Results are written here during the running process.

    """

    def __init__(self, pipe,
                 optimization_info,
                 outer_fold_id,
                 cross_validation_info,
                 cache_folder=None,
                 cache_updater=None,
                 dummy_estimator=None,
                 result_obj=None):
        self.outer_fold_id = outer_fold_id
        self.cross_validation_info = cross_validation_info
        self.optimization_info = optimization_info
        self._pipe = pipe
        self.copy_pipe_fnc = self._pipe.copy_me
        self.dummy_estimator = dummy_estimator

        self.cache_folder = cache_folder
        self.cache_updater = cache_updater

        # Information about the optimization progress
        self.current_best_config = None
        self.optimizer = None
        self.constraint_objects = None

        # data
        self.result_object = result_obj
        self.inner_folds = None
        self._validation_X = None
        self._validation_y = None
        self._validation_kwargs = None
        self._test_X = None
        self._test_y = None
        self._test_kwargs = None

    def _prepare_optimization(self):

        logger.info("Preparing Hyperparameter Optimization...")
        pipeline_elements = [e for name, e in self._pipe.elements]

        self.optimizer = self.optimization_info.get_optimizer()
        if isinstance(self.optimizer, PhotonMasterOptimizer):
            self.optimizer.prepare(pipeline_elements, self.optimization_info.maximize_metric, self.objective_function)
        else:
            self.optimizer.prepare(pipeline_elements, self.optimization_info.maximize_metric)

        # we've got some super strange pymodm problems here
        # somehow some information from the previous outer fold lingers on and can be found within a completely new
        # instantiated OuterFoldMDB object
        # hence, clearing it
        self.result_object.tested_config_list = list()

        # copy constraint objects.
        if self.optimization_info.performance_constraints is not None:
            if isinstance(self.optimization_info.performance_constraints, list):
                self.constraint_objects = [original.copy_me() for original in self.optimization_info.performance_constraints]
            else:
                self.constraint_objects = [self.optimization_info.performance_constraints.copy_me()]
        else:
            self.constraint_objects = None

    def _prepare_data(self, X, y=None, **kwargs):
        logger.info("Preparing data for outer fold " + str(self.cross_validation_info.outer_folds[self.outer_fold_id].fold_nr) + "...")
        # Prepare Train and validation set data
        train_indices = self.cross_validation_info.outer_folds[self.outer_fold_id].train_indices
        test_indices = self.cross_validation_info.outer_folds[self.outer_fold_id].test_indices
        self._validation_X, self._validation_y, self._validation_kwargs = PhotonDataHelper.split_data(X, y, kwargs,
                                                                                                      indices=train_indices)
        self._test_X, self._test_y, self._test_kwargs = PhotonDataHelper.split_data(X, y, kwargs, indices=test_indices)

        # write numbers to database info object
        self.result_object.number_samples_validation = self._validation_y.shape[0]
        self.result_object.number_samples_test = self._test_y.shape[0]
        if self._pipe._estimator_type == "classifier":
            self.result_object.class_distribution_validation = FoldInfo.data_overview(self._validation_y)
            self.result_object.class_distribution_test = FoldInfo.data_overview(self._test_y)

    def _generate_inner_folds(self):

        self.inner_folds = FoldInfo.generate_folds(self.cross_validation_info.inner_cv,
                                                   self._validation_X,
                                                   self._validation_y,
                                                   self._validation_kwargs)

        self.cross_validation_info.inner_folds[self.outer_fold_id] = {f.fold_id: f for f in self.inner_folds}

    def fit(self, X, y=None, **kwargs):
        logger.photon_system_log('')
        logger.stars()
        logger.photon_system_log('Outer Cross validation Fold {}'.format(self.cross_validation_info.outer_folds[self.outer_fold_id].fold_nr))
        logger.stars()

        self._prepare_data(X, y, **kwargs)
        self._prepare_optimization()
        self._fit_dummy()
        self._generate_inner_folds()

        outer_fold_fit_start_time = datetime.datetime.now()
        self.best_metric_yet = None
        self.tested_config_counter = 0

        # distribute number of folds to encapsulated child hyperpipes
        # self.__distribute_cv_info_to_hyperpipe_children(num_of_folds=num_folds,
        #                                                 outer_fold_counter=outer_fold_counter)

        if self.cross_validation_info.calculate_metrics_per_fold:
            self.fold_operation = "mean"
        else:
            self.fold_operation = "raw"

        self.max_nr_of_configs = ''
        if hasattr(self.optimizer, 'n_configurations'):
            self.max_nr_of_configs = str(self.optimizer.n_configurations)

        if isinstance(self.optimizer, PhotonMasterOptimizer):
            self.optimizer.optimize()
        else:
            # do the optimizing
            for current_config in self.optimizer.ask:
                self.objective_function(current_config)

        logger.line()
        logger.info('Hyperparameter Optimization finished. Now finding best configuration .... ')
        logger.info(self.tested_config_counter)
        # now go on with the best config found
        if self.tested_config_counter > 0:
            best_config_outer_fold = self.result_object.get_optimum_config(self.optimization_info.best_config_metric,
                                                                           self.optimization_info.maximize_metric,
                                                                           fold_operation=self.fold_operation)
            # inform user
            logger.debug('Optimizer metric: ' + self.optimization_info.best_config_metric + '\n' +
                         '   --> Maximize metric: ' + str(self.optimization_info.maximize_metric))

            logger.system_line()
            logger.photon_system_log('BEST_CONFIG ')
            logger.system_line()
            logger.photon_system_log(json.dumps(best_config_outer_fold.human_readable_config, indent=4,
                                                sort_keys=True))
            logger.system_line()
            logger.photon_system_log('VALIDATION PERFORMANCE')
            logger.system_line()
            print_double_metrics(best_config_outer_fold.get_train_metric(operation="mean"),
                                 best_config_outer_fold.get_test_metric(operation="mean"))

            if not best_config_outer_fold:
                raise Exception("No best config was found!")

            # ... and create optimal pipeline
            optimum_pipe = self.copy_pipe_fnc()
            if self.cache_updater is not None:
                self.cache_updater(optimum_pipe, self.cache_folder, "fixed_fold_id")
            optimum_pipe.caching = False
            # set self to best config
            optimum_pipe.set_params(**best_config_outer_fold.config_dict)

            # Todo: set all children to best config and inform to NOT optimize again, ONLY fit
            # for child_name, child_config in best_config_outer_fold_mdb.children_config_dict.items():
            #     if child_config:
            #         # in case we have a pipeline stacking we need to identify the particular subhyperpipe
            #         splitted_name = child_name.split('__')
            #         if len(splitted_name) > 1:
            #             stacking_element = self.optimum_pipe.named_steps[splitted_name[0]]
            #             pipe_element = stacking_element.elements[splitted_name[1]]
            #         else:
            #             pipe_element = self.optimum_pipe.named_steps[child_name]
            #         pipe_element.set_params(**child_config)
            #         pipe_element.is_final_fit = True

            # self.__distribute_cv_info_to_hyperpipe_children(reset=True)

            logger.debug('Fitting model with best configuration of outer fold...')
            optimum_pipe.fit(self._validation_X, self._validation_y, **self._validation_kwargs)

            self.result_object.best_config = best_config_outer_fold

            # save test performance
            best_config_performance_mdb = MDBInnerFold()
            best_config_performance_mdb.fold_nr = -99
            best_config_performance_mdb.number_samples_training = self._validation_y.shape[0]
            best_config_performance_mdb.number_samples_validation = self._test_y.shape[0]
            best_config_performance_mdb.feature_importances = optimum_pipe.feature_importances_

            if self.cross_validation_info.use_test_set:
                # Todo: generate mean and std over outer folds as well. move this items to the top
                logger.info('Calculating best model performance on test set...')

                logger.debug('...scoring test data')
                test_score_mdb = InnerFoldManager.score(optimum_pipe, self._test_X, self._test_y,
                                                        indices=self.cross_validation_info.outer_folds[self.outer_fold_id].test_indices,
                                                        metrics=self.optimization_info.metrics,
                                                        **self._test_kwargs)

                logger.debug('... scoring training data')

                train_score_mdb = InnerFoldManager.score(optimum_pipe, self._validation_X, self._validation_y,
                                                         indices=self.cross_validation_info.outer_folds[self.outer_fold_id].train_indices,
                                                         metrics=self.optimization_info.metrics,
                                                         training=True,
                                                         **self._validation_kwargs)

                best_config_performance_mdb.training = train_score_mdb
                best_config_performance_mdb.validation = test_score_mdb

                logger.system_line()
                logger.photon_system_log('TEST PERFORMANCE')
                logger.system_line()
                print_double_metrics(train_score_mdb.metrics, test_score_mdb.metrics)
            else:

                def _copy_inner_fold_means(metric_dict):
                    # We copy all mean values from validation to the best config
                    # training
                    train_item_metrics = {}
                    for m in metric_dict:
                        if m.operation == str(self.fold_operation):
                            train_item_metrics[m.metric_name] = m.value
                    train_item = MDBScoreInformation()
                    train_item.metrics_copied_from_inner = True
                    train_item.metrics = train_item_metrics
                    return train_item

                # training
                best_config_performance_mdb.training = _copy_inner_fold_means(best_config_outer_fold.metrics_train)
                # validation
                best_config_performance_mdb.validation = _copy_inner_fold_means(best_config_outer_fold.metrics_test)

            # write best config performance to best config item
            self.result_object.best_config.best_config_score = best_config_performance_mdb

        logger.info('Computations in outer fold {} took {} minutes.'.format(
            self.cross_validation_info.outer_folds[self.outer_fold_id].fold_nr,
            (datetime.datetime.now() - outer_fold_fit_start_time).total_seconds() / 60))

    def objective_function(self, current_config):
        if current_config is None:
            return
        logger.line()
        self.tested_config_counter += 1

        if hasattr(self.optimizer, 'ask_for_pipe'):
            pipe_ctor = self.optimizer.ask_for_pipe()
        else:
            pipe_ctor = self.copy_pipe_fnc

        # self.__distribute_cv_info_to_hyperpipe_children(reset=True, config_counter=tested_config_counter)

        hp = InnerFoldManager(pipe_ctor, current_config,
                              self.optimization_info,
                              self.cross_validation_info, self.outer_fold_id, self.constraint_objects,
                              cache_folder=self.cache_folder,
                              cache_updater=self.cache_updater)

        # Test the configuration cross validated by inner_cv object
        current_config_mdb = hp.fit(self._validation_X, self._validation_y, **self._validation_kwargs)
        current_config_mdb.config_nr = self.tested_config_counter

        if not current_config_mdb.config_failed:
            metric_train = current_config_mdb.get_train_metric(self.optimization_info.best_config_metric,
                                                               self.fold_operation)
            metric_test = current_config_mdb.get_test_metric(self.optimization_info.best_config_metric,
                                                             self.fold_operation)

            if metric_train is None or metric_test is None:
                raise Exception("Config did not fail, but did not get any metrics either....!!?")
            config_performance = (metric_train, metric_test)
            if self.best_metric_yet is None:
                self.best_metric_yet = config_performance
                self.current_best_config = current_config_mdb
            else:
                # check if we have the next superstar around that exceeds any old performance
                if self.optimization_info.maximize_metric:
                    if metric_test > self.best_metric_yet[1]:
                        self.best_metric_yet = config_performance
                        self.current_best_config.decrease_memory()
                        self.current_best_config = current_config_mdb
                    else:
                        current_config_mdb.decrease_memory()
                else:
                    if metric_test < self.best_metric_yet[1]:
                        self.best_metric_yet = config_performance
                        self.current_best_config.decrease_memory()
                        self.current_best_config = current_config_mdb
                    else:
                        current_config_mdb.decrease_memory()

            # Print Result for config
            computation_duration = current_config_mdb.computation_end_time - current_config_mdb.computation_start_time
            logger.info('Computed configuration ' + str(self.tested_config_counter) + "/" + self.max_nr_of_configs +
                        " in " + str(computation_duration))
            logger.info("Performance:             " + self.optimization_info.best_config_metric
                        + " - Train: " + "%.4f" % config_performance[0] + ", Validation: " + "%.4f" %
                        config_performance[1])
            logger.info("Best Performance So Far: " + self.optimization_info.best_config_metric
                        + " - Train: " + "%.4f" % self.best_metric_yet[0] + ", Validation: "
                        + "%.4f" % self.best_metric_yet[1])
        else:
            config_performance = (-1, -1)
            # Print Result for config
            logger.debug('...failed:')
            logger.error(current_config_mdb.config_error)

        # add config to result tree
        self.result_object.tested_config_list.append(current_config_mdb)

        # 3. inform optimizer about performance
        logger.debug("Telling hyperparameter optimizer about recent performance.")
        if isinstance(self.optimizer, PhotonSlaveOptimizer):
            self.optimizer.tell(current_config, config_performance[1])
        logger.debug("Asking hyperparameter optimizer for new config.")

        if self.optimization_info.maximize_metric:
            return 1 - config_performance[1]
        else:
            return config_performance[1]

    def _fit_dummy(self):
        if self.dummy_estimator is not None:
            logger.info("Running Dummy Estimator...")
            try:
                if isinstance(self._validation_X, np.ndarray):
                    if len(self._validation_X.shape) > 2:
                        logger.info("Skipping dummy estimator because of too many dimensions")
                        self.result_object.dummy_results = None
                        return
                dummy_y = np.reshape(self._validation_y, (-1, 1))
                self.dummy_estimator.fit(dummy_y, self._validation_y)
                train_scores = InnerFoldManager.score(self.dummy_estimator, self._validation_X, self._validation_y,
                                                      metrics=self.optimization_info.metrics)

                # fill result tree with fold information
                inner_fold = MDBInnerFold()
                inner_fold.training = train_scores

                if self.cross_validation_info.use_test_set:
                    test_scores = InnerFoldManager.score(self.dummy_estimator,
                                                         self._test_X, self._test_y,
                                                         metrics=self.optimization_info.metrics)
                    print_metrics("DUMMY", test_scores.metrics)
                    inner_fold.validation = test_scores

                self.result_object.dummy_results = inner_fold

                # performaceConstraints: DummyEstimator
                if self.constraint_objects is not None:
                    dummy_constraint_objs = [opt for opt in self.constraint_objects
                                             if isinstance(opt, DummyPerformanceConstraint)]

                    if dummy_constraint_objs:
                        for dummy_constraint_obj in dummy_constraint_objs:
                            dummy_constraint_obj.set_dummy_performance(self.result_object.dummy_results)

                return inner_fold
            except Exception as e:
                logger.error(e)
                logger.info("Skipping dummy because of error..")
                return None
        else:
            logger.info("Skipping dummy ..")



from sklearn.base import BaseEstimator


class CustomCovariateTransformer(BaseEstimator):
    """
    Any algorithm that uses the target values to change the data (e.g. transform male and female differently?)
    or that changes the dataset including targets in total (e.g. sample pairing oversampling)
    """

    def __init__(self, param1=0, param2=None):
        # it is important that you name your params the same in the constructor
        #  stub as well as in your class variables!
        self.param1 = param1
        self.param2 = param2

        self.needs_covariates = True

    def fit(self, data, targets=None, **kwargs):
        """
        Adjust the underlying model or method to the data.
        You can find any covariate data in the kwargs.

        E.g. if you did hyperpipe.fit(X, y, age=my_age_variable)
        you can access the matched age values to X and y by
        age = kwargs["age"]
        Returns
        -------
        IMPORTANT: must return self!
        """
        return self

    def transform(self, data, targets=None, **kwargs):
        """
        Apply the method's logic to the data using any covariate not included in the feature set X.

        E.g. if you did hyperpipe.fit(X, y, age=my_age_variable)
        you can access the matched age values to X and y by
        age = kwargs["age"]

        You can change the  values for later applied algorithms by modifying them and returning a new dict
        kwargs["age"] = new_age_variable

        IMPORTANT: MUST return both data and kwargs
        """
        return data, kwargs



import uuid
import numpy as np
from bson.objectid import ObjectId
from sklearn.datasets import load_breast_cancer
from sklearn.model_selection import KFold

from photonai.base import Hyperpipe, OutputSettings, PipelineElement
from photonai.processing.permutation_test import PermutationTest
from photonai.processing.results_handler import ResultsHandler
from photonai.helper.photon_base_test import PhotonBaseTest


class PermutationTestTests(PhotonBaseTest):

    @classmethod
    def setUpClass(cls) -> None:
        cls.file = __file__
        super(PermutationTestTests, cls).setUpClass()

        cls.perm_id = uuid.uuid4()
        cls.wizard_obj_id = ObjectId()
        cls.hyperpipe = Hyperpipe("permutation_test_pipe",
                                  inner_cv=KFold(n_splits=4),
                                  outer_cv=KFold(n_splits=3),
                                  metrics=["accuracy", "balanced_accuracy"],
                                  best_config_metric="balanced_accuracy",
                                  output_settings=OutputSettings(
                                      mongodb_connect_url="mongodb://localhost:27017/photon_results",
                                      wizard_object_id=str(cls.wizard_obj_id)),
                                  project_folder=cls.tmp_folder_path,
                                  verbosity=0,
                                  permutation_id=str(cls.perm_id) + "_reference")
        cls.hyperpipe += PipelineElement("StandardScaler")
        cls.hyperpipe += PipelineElement("SVC")
        cls.X, cls.y = load_breast_cancer(return_X_y=True)
        cls.hyperpipe.fit(cls.X, cls.y)

    def test_wizard_preparation(self):
        result = PermutationTest.prepare_for_wizard(str(self.perm_id), self.wizard_obj_id,
                                                    mongo_db_connect_url="mongodb://localhost:27017/photon_results")
        returned_duration = result["estimated_duration"]
        cached_duration = self.hyperpipe.results.computation_end_time - self.hyperpipe.results.computation_start_time
        self.assertAlmostEqual(round(returned_duration.total_seconds(), 3),
                               round(cached_duration.total_seconds(), 3), 2)
        # Todo: setup case where it is false
        self.assertTrue(result["usability"])

        my_handler = ResultsHandler()
        my_handler.load_from_mongodb(self.hyperpipe.output_settings.mongodb_connect_url, str(self.wizard_obj_id))
        self.assertEqual(my_handler.results.permutation_id,
                         PermutationTest.get_mother_permutation_id(str(self.perm_id)))

    def test_find_reference(self):
        obj_id = self.hyperpipe.results._id
        returned_obj_id = PermutationTest.find_reference(self.hyperpipe.output_settings.mongodb_connect_url,
                                                         str(self.perm_id), False)._id
        self.assertEqual(obj_id, returned_obj_id)

        wizard_obj_id = str(self.wizard_obj_id)
        latest_item = PermutationTest.find_reference(self.hyperpipe.output_settings.mongodb_connect_url,
                                                     ObjectId(wizard_obj_id), True)
        self.assertEqual(latest_item.name, wizard_obj_id)

    def create_hyperpipe(self):
        # this is needed here for the parallelisation
        from photonai.base import Hyperpipe, PipelineElement, OutputSettings
        from photonai.optimization import IntegerRange
        from sklearn.model_selection import GroupKFold
        from sklearn.model_selection import KFold

        settings = OutputSettings(mongodb_connect_url='mongodb://localhost:27017/photon_results')
        my_pipe = Hyperpipe('permutation_test_1',
                            optimizer='grid_search',
                            metrics=['accuracy', 'precision', 'recall'],
                            best_config_metric='accuracy',
                            outer_cv=GroupKFold(n_splits=2),
                            inner_cv=KFold(n_splits=2),
                            calculate_metrics_across_folds=True,
                            use_test_set=True,
                            verbosity=0,
                            project_folder=self.tmp_folder_path,
                            output_settings=settings)

        # Add transformer elements
        my_pipe += PipelineElement("StandardScaler", hyperparameters={},
                                   test_disabled=False, with_mean=True, with_std=True)

        my_pipe += PipelineElement("PCA", hyperparameters={'n_components': IntegerRange(3, 5)},
                                   test_disabled=False)

        # Add estimator
        my_pipe += PipelineElement("SVC", hyperparameters={'kernel': ['linear', 'rbf']},  # C': FloatRange(0.1, 5),
                                   gamma='scale', max_iter=1000000)

        return my_pipe

    def create_hyperpipe_no_mongo(self):
        from photonai.base import Hyperpipe
        from sklearn.model_selection import KFold

        my_pipe = Hyperpipe('permutation_test_1',
                            optimizer='grid_search',
                            metrics=['accuracy', 'precision', 'recall'],
                            best_config_metric='accuracy',
                            outer_cv=KFold(n_splits=2),
                            inner_cv=KFold(n_splits=2),
                            calculate_metrics_across_folds=True,
                            use_test_set=True,
                            verbosity=0,
                            project_folder=self.tmp_folder_path)
        return my_pipe

    def test_no_mongo_connection_string(self):
        perm_tester = PermutationTest(self.create_hyperpipe_no_mongo, n_perms=2, n_processes=3, random_state=11,
                                      permutation_id=str(uuid.uuid4()))
        with self.assertRaises(ValueError):
            perm_tester.fit(self.X, self.y)

    def test_run_parallelized_perm_test(self):
        X, y = load_breast_cancer(return_X_y=True)
        my_perm_id = str(uuid.uuid4())
        groups = np.random.random_integers(0, 3, (len(y),))
        perm_tester = PermutationTest(self.create_hyperpipe, n_perms=2, n_processes=3, random_state=11,
                                      permutation_id=my_perm_id)
        perm_tester.fit(X, y, groups=groups)

    def test_setup_non_useful_perm_test(self):
        np.random.seed(1335)
        X, y = np.random.random((200, 5)), np.random.randint(0, 2, size=(200, ))
        my_perm_id = str(uuid.uuid4())
        groups = np.random.random_integers(0, 3, (len(y),))
        perm_tester = PermutationTest(self.create_hyperpipe, n_perms=2, n_processes=3, random_state=11,
                                      permutation_id=my_perm_id)
        with self.assertRaises(RuntimeError):
            perm_tester.fit(X, y, groups=groups)

    def test_run_perm_test(self):
        X, y = load_breast_cancer(return_X_y=True)
        my_perm_id = str(uuid.uuid4())
        groups = np.random.random_integers(0, 3, (len(y),))
        perm_tester = PermutationTest(self.create_hyperpipe, n_perms=2, n_processes=1, random_state=11,
                                      permutation_id=my_perm_id)
        perm_tester.fit(X, y, groups=groups)

        results = PermutationTest._calculate_results(my_perm_id,
                                                     mongodb_path='mongodb://localhost:27017/photon_results')

        self.assertAlmostEqual(results.p_values['accuracy'], 0)



from sklearn.datasets import load_breast_cancer
from sklearn.model_selection import KFold

from photonai.base import Hyperpipe, PipelineElement, Stack, Branch
from photonai.optimization import IntegerRange, Categorical, FloatRange

X, y = load_breast_cancer(return_X_y=True)

my_pipe = Hyperpipe('basic_stacking',
                    optimizer='grid_search',
                    metrics=['accuracy', 'precision', 'recall'],
                    best_config_metric='f1_score',
                    outer_cv=KFold(n_splits=3),
                    inner_cv=KFold(n_splits=10),
                    verbosity=1,
                    project_folder='./tmp/')

# BRANCH WITH QUANTILTRANSFORMER AND DECISIONTREECLASSIFIER
tree_qua_branch = Branch('tree_branch')
tree_qua_branch += PipelineElement('QuantileTransformer', n_quantiles=100)
tree_qua_branch += PipelineElement('DecisionTreeClassifier',
                                   {'min_samples_split': IntegerRange(2, 4)},
                                   criterion='gini')

# BRANCH WITH MinMaxScaler AND DecisionTreeClassifier
svm_mima_branch = Branch('svm_branch')
svm_mima_branch += PipelineElement('MinMaxScaler')
svm_mima_branch += PipelineElement('SVC',
                                   {'kernel': Categorical(['rbf', 'linear']),
                                    'C': FloatRange(0.01, 2.0, num=10)},
                                   gamma='auto')

# BRANCH WITH StandardScaler AND KNeighborsClassifier
knn_sta_branch = Branch('neighbour_branch')
knn_sta_branch += PipelineElement('StandardScaler')
knn_sta_branch += PipelineElement('KNeighborsClassifier')

# voting = True to mean the result of every branch
my_pipe += Stack('final_stack', [tree_qua_branch, svm_mima_branch, knn_sta_branch])

my_pipe += PipelineElement('LogisticRegression', solver='lbfgs')

my_pipe.fit(X, y)



from typing import Union
import numpy as np
from keras.optimizers import Optimizer
import photonai.modelwrapper.keras_base_models as keras_dnn_base_model

from photonai.modelwrapper.keras_base_models import KerasDnnBaseModel, KerasBaseRegressor
from photonai.photonlogger import logger


class KerasDnnRegressor(KerasDnnBaseModel, KerasBaseRegressor):
    """Wrapper class for a regression-based Keras model.

    See [Keras API](https://keras.io/api/).

    Example:
        ``` python
        PipelineElement('KerasDnnRegressor',
                        hyperparameters={'hidden_layer_sizes': Categorical([[18, 14], [30, 5]]),
                                         'dropout_rate': Categorical([0.01, 0.2])},
                        activations='relu',
                        epochs=50,
                        nn_batch_size=64,
                        verbosity=1)
        ```

    """
    def __init__(self,
                 hidden_layer_sizes: int = None,
                 learning_rate: float = 0.01,
                 loss: str = "mean_squared_error",
                 epochs: int = 10,
                 nn_batch_size: int = 64,
                 metrics: list = None,
                 validation_split: float = 0.1,
                 callbacks: list = None,
                 batch_normalization: bool = True,
                 verbosity: int = 0,
                 dropout_rate: Union[float, list] = 0.2,
                 activations: Union[str, list] = 'relu',
                 optimizer: Union[Optimizer, str] = "adam"):
        """
        Initialize the object.

        Parameters:
            hidden_layer_sizes:
                Number of perceptrons per layer.

            learning_rate:
                Step size of the learning adjustment.

            loss:
                Loss function.

            epochs:
                Number of arbitrary cutoffs, generally defined as
                "one pass over the entire dataset", used to separate training into distinct phases,
                which is useful for logging and periodic evaluation.

            nn_batch_size:
                Typically the batch_size. A batch is a set of nn_batch_size samples.
                The samples in a batch are processed independently, in parallel.
                If training, a batch results in only one update to the model.

            metrics:
                List of evaluate metrics.

            callbacks:
                Within Keras, there is the ability to add callbacks specifically designed
                to be run at the end of an epoch. Examples of these
                are learning rate changes and model checkpointing (saving).

            validation_split:
                Split size of validation set.

            batch_normalization:
                Batch normalization applies a transformation that maintains
                the mean output close to 0 and the output standard deviation close to 1.

            verbosity:
                The level of verbosity, 0 is least talkative and
                gives only warn and error, 1 gives adds info and 2 adds debug.

            dropout_rate:
                A Dropout layer applies random dropout and rescales the output.
                In inference mode, the same layer does nothing.
                Float -> added behind each layer
                List -> Same size as hidden_layer_size

            activations:
                Activation function.

            optimizer:
                Optimization algorithm.

        """
        self._loss = ""
        self._multi_class = None
        self.loss = loss
        self.epochs = epochs
        self.nn_batch_size = nn_batch_size
        self.validation_split = validation_split

        if callbacks:
            self.callbacks = callbacks
        else:
            self.callbacks = []

        if not metrics:
            metrics = ['mean_squared_error']

        super(KerasDnnRegressor, self).__init__(hidden_layer_sizes=hidden_layer_sizes,
                                                target_activation="linear",
                                                target_dimension=1,
                                                learning_rate=learning_rate,
                                                loss=loss,
                                                metrics=metrics,
                                                batch_normalization=batch_normalization,
                                                verbosity=verbosity,
                                                dropout_rate=dropout_rate,
                                                activations=activations,
                                                optimizer=optimizer)

    @property
    def target_activation(self):
        return "linear"

    @target_activation.setter
    def target_activation(self, value):
        if value != "linear":
            msg = "The subclass of KerasBaseRegressor does not allow to use another " \
                  "target_activation. Please use 'linear' like default."
            logger.error(msg)
            raise ValueError(msg)

    @property
    def loss(self):
        return self._loss

    @loss.setter
    def loss(self, value):
        if value in keras_dnn_base_model.get_loss_allocation()["regression"]:
            self._loss = value
        else:
            raise ValueError("Loss function is not supported. Feel free to use upperclass without restrictions.")

    def fit(self, X: np.ndarray, y: np.ndarray):
        """
        Starting the learning.

        Parameters:
            X:
                The input samples with shape [n_samples, n_features].

            y:
                The input targets with shape [n_samples, 1].

        """
        self.create_model(X.shape[1])
        super(KerasDnnBaseModel, self).fit(X, y)
        return self



from sklearn.datasets import load_breast_cancer
from sklearn.model_selection import KFold

from photonai.base import Hyperpipe, PipelineElement, Preprocessing, Stack
from photonai.optimization import FloatRange, Categorical

# WE USE THE BREAST CANCER SET FROM SKLEARN
X, y = load_breast_cancer(return_X_y=True)

# DESIGN YOUR PIPELINE
my_pipe = Hyperpipe('svm_ensemble_pipe',
                    optimizer='sk_opt',
                    optimizer_params={'n_configurations': 10},
                    metrics=['accuracy', 'precision', 'recall', 'balanced_accuracy'],
                    best_config_metric='accuracy',
                    outer_cv=KFold(n_splits=3),
                    inner_cv=KFold(n_splits=3),
                    project_folder='./tmp/')


preprocessing = Preprocessing()
preprocessing += PipelineElement("LabelEncoder")
my_pipe += preprocessing

my_pipe.add(PipelineElement('StandardScaler'))

my_stack = Stack("SVC_Ensemble")
for i in range(20):
    my_stack += PipelineElement('SVC', hyperparameters={'kernel': Categorical(['rbf', 'linear']),
                                                        'C': FloatRange(0.5, 2, step=0.5)}, gamma='scale')
my_pipe += my_stack
my_pipe += PipelineElement('GaussianProcessClassifier')

my_pipe.fit(X, y)
